#include <iostream>
#include <string>

#include "risc.h"

#include <libgen.h>

#define HELP_PAGE                                                                   \
"elab: " RISC_LIBRARY_NAME " Dynamic Elaborator, V" RISC_VERSION_NUMBER "\n"        \
"      " RISC_COPYRIGHT "\n"                                                        \
"\n"                                                                                \
"Usage: elab [<option>...] <design> [<option>...]\n"                                \
"\n"                                                                                \
"Options:               Description:                    Default:\n"                 \
"-h|--help              print this usage info, quit     none\n"                     \
"-v|--verbose           increment verbosity             silent\n"                   \
"-vv                    increment verbosity by 2        silent\n"                   \
"-vvv                   increment verbosity by 3        silent\n"                   \
"-w|--warnings          increment warning level         no warnings\n"              \
"-ww                    increment warning level by 2    no warnings\n"              \
"-www                   increment warning level by 3    no warnings\n"              \
"-g                     produce debugging information   no debugging infos\n"       \
"-O|-O<level>           enable optimizations            no optimization\n"          \
"-I<dir>                add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-L<dir>                add dir to library path         $SYSTEMC_OOP_HOME/lib\n"    \
"-l<lib>                add lib to linker libraries     -lsystemc\n"                \
"-c                     compile only, do not link       compile and link\n"         \
"-o <filename>          name output file as filename    a.out\n"                    \
"-elab:o <filename>     export elaboration result       a.elab\n"                   \
"<rose:option>          passed to ROSE compiler         none\n"                     \
"<g++_option>           passed to GNU C++ compiler      none\n"                     \
"\n"

class RISCParseStatus: public risc::tools::ParseStatus
{
  public:

    std::string filename_elab_results;
    std::string filename_executable;

    RISCParseStatus():
      ParseStatus()
    { }

    RISCParseStatus& operator=(const RISCParseStatus& copy)
    {
      verbosityLevel = copy.verbosityLevel;
      warningLevel = copy.warningLevel;
      compilerName = copy.compilerName;

      return *this;
    }
};

void
RISCParseCommandlineOptions(int &argc, char** &argv, RISCParseStatus* status)
{
  if (argc < 2)
  {
    std::cout << "elab: no input files\n";
    exit(0);
  }

  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    // -h, --help
    if (it->compare("-h") == 0 || it->compare("--help") == 0)
    {
      std::cout << HELP_PAGE;
      exit(0);
    }

    if (it->compare("-elab:o") == 0)
    {
      RemovedIter.push_back(it);
      it++;
      RemovedIter.push_back(it);
      if(it == argList.end()) {
        std::cout << "Flag -elab:o has no filename" << std::endl;
        exit(1);
      }
      status->filename_elab_results = *it;
    }

    if (it->compare("-o") == 0)
    {
      it++;
      if(it == argList.end()) {
        std::cout << "Flag -o has no filename" << std::endl;
        exit(1);
      }
      status->filename_executable = *it;
    }

  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);

  // Parse library-level command-line options
  risc::tools::ParseCommandlineOptions(argc, argv, status);
}


int main(int argc, char* argv[])
{
  RISCParseStatus status;

  RISCParseCommandlineOptions(argc, argv, &status);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "ELAB starting..."
              << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "ELAB: adding command-line options for frontend..."
              << std::endl;
  }

  // add -DRISC (RD, 07/06/18)
  risc::tools::AddCommandlineOption(argc, argv, "-DRISC");

  std::string LWIncludePath;
  std::string LWLibraryPath;

  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";

  std::string LWLibraryPathwoL = LWLibraryPath;
  LWLibraryPathwoL.erase(0, 2);

  // add -I$SYSTEMC_LW_HOME/include
  risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);

  // add -L$SYSTEMC_LW_HOME/lib-linux64
  risc::tools::AddCommandlineOption(argc, argv, LWLibraryPath);

  // add -lsystemc
  risc::tools::AddCommandlineOption(argc, argv, "-lsystemc");


  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "ELAB: parsing, building ROSE IR..." << std::endl;
  }
  risc::Design d(argc, argv, std::vector<std::string>(0), false,
                  status.verbosityLevel >= RISC_VERY_VERY_VERBOSE);

  // add -Xlinker -R -Xlinker $SYSTEMC_LW_HOME/lib-linux64
  risc::tools::AddCommandlineOption(d, "-Xlinker");
  risc::tools::AddCommandlineOption(d, "-R");
  risc::tools::AddCommandlineOption(d, "-Xlinker");
  risc::tools::AddCommandlineOption(d, LWLibraryPathwoL);

  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "ELAB: instrumenting..." << std::endl;
  }

  if (status.filename_elab_results.empty()) {

    risc::inst::instrument_reference_printing("a.elab");
  } else {
    risc::inst::instrument_reference_printing(status.filename_elab_results);
  }

  risc::inst::risc_unparse(d, &status, "ELAB", NULL, NULL);

/*** THIS PATCHING AROUND ROSE LIMITATIONS IS NOT NEEDED ANY MORE
     WITH THE NEW ROSE LIBRARY VERSION 0.9.10.X (RD, 05/30/18)
  risc::inst::risc_patch(d, &status,
      "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_method_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
      "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
      "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
      "(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
      "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
      "\\5)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
      "ELAB");
***/

  risc::inst::risc_compile(d, &status, "ELAB");

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
