// ----------------------------------------------------------------------------
// visual.cpp
// ----------------------------------------------------------------------------
//
// Author: Daniel Mendoza
//
// 
// ----------------------------------------------------------------------------


#include <gtk/gtk.h>
#include <cairo.h>
#include <stdlib.h>
#include <time.h>
#include <sstream>
#include <map>
#include "risc.h"
#include "../../../src/tools/function_helper.h"

#define D_SPACE 20 //needed for module sizing
#define MINW 40
#define MINH 50

#define SOCKET_SIZE 12

#define THREAD_SIZE 20

#define TEXT_CHAR 8

#define DOT_WIDTH 8

#define TRANS_VAR 0.2

#define PRIM_CHANNEL_DISTANCE 20
#define PORT_TEXT_XOFFSET 5
#define PORT_TEXT_YOFFSET 10
#define COLOR_PALET_FORMATTER 13

const char* font_name = "Purisa";

struct pair_comp{
	bool operator()(const std::pair<int, SgVariableDefinition*>& lhs, 
					const std::pair<int, SgVariableDefinition*>& rhs) const
	{
		std::string name_lhs = lhs.second->get_declaration()->get_decl_item(NULL)->get_qualified_name();
		std::string key_lhs = std::to_string(static_cast<long long>(lhs.first)) + " " + name_lhs;
		std::string name_rhs = rhs.second->get_declaration()->get_decl_item(NULL)->get_qualified_name();
		std::string key_rhs = std::to_string(static_cast<long long>(rhs.first)) + " " + name_rhs;
		//std::cout << "comparing : " << key_lhs
		//	<< "    against     " << key_rhs << std::endl;
		return key_lhs < key_rhs;
	}
};
typedef std::map<std::pair<int, SgVariableDefinition*> , risc::InstanceTree, pair_comp> OrderedTreeMap;

OrderedTreeMap ordered_tree_mapping;
/*
 * Visual Additional IR
 *
 */

struct VisualInitSocket {
	risc::InitiatorSocket* ptr;
	int instance_id;
	bool HB;
	bool HB_connected;

	VisualInitSocket(risc::InitiatorSocket* p, int id) {
		ptr = p;
		instance_id = id;
		HB = false;
		HB_connected = false;
	}
};

struct VisualTargSocket {
	risc::TargetSocket* ptr;
	int instance_id;
	bool HB;
	bool HB_connected;

	VisualTargSocket(risc::TargetSocket* p, int id) {
		ptr = p;
		instance_id = id;
		HB = false;
		HB_connected = false;
	}
};

struct VisualDrawnSocket {
	int x;
	int y;
	SgSymbol* target;
	int target_instance_id;
	bool initiator;

	VisualDrawnSocket() {
	}

	VisualDrawnSocket(int in_x, int in_y, SgSymbol* in_targ, int in_id,bool init) {
		x = in_x;
		y = in_y;
		target = in_targ;
		target_instance_id = in_id;	
		initiator = init;
	}
};

struct pairhash {
public:
  template <typename T, typename U>
  std::size_t operator()(const std::pair<T, U> &x) const
  {
    return std::hash<T>()(x.first) ^ std::hash<U>()(x.second);
  }
};

struct VisualModule {
	risc::Instance *mod;
	std::vector<VisualModule*> sub_modules;
	std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash> sub_mod_HB_map;

	std::unordered_map<std::pair<int,SgSymbol*>, int, pairhash> sub_mod_channel_map;

	std::unordered_map<std::pair<int,SgSymbol*> ,int,pairhash> HB_init_sockets;
	std::unordered_map<std::pair<int,SgSymbol*> ,int,pairhash> HB_targ_sockets;
	
	std::vector<VisualInitSocket*> initiator_sockets;
	std::vector<VisualTargSocket*> target_sockets;

	int instance_id;

	VisualModule(risc::Instance* m) {
		mod = m;
		std::list<risc::InstanceTree> temp2 = risc::get_design()->get_root().find_all_instances_of(m->get_symbol());
		int cur_mod_id = -1;
		for (std::list<risc::InstanceTree>::iterator list_it = temp2.begin();
			list_it != temp2.end();
			list_it++) {
					if(list_it->get_object()->get_name() == m->get_name()) {	
						cur_mod_id = risc::get_design()->path_instance_mapper_->get_id(*list_it);
						break;
					}
		}
		if(cur_mod_id == -1) {
			printf("could not find current module id!\n");
		}
		instance_id = cur_mod_id;
		for(risc::InitiatorSocketVectorIter init_iter = m->get_definition()->initiatorsockets_.begin();
			init_iter != m->get_definition()->initiatorsockets_.end();
			init_iter++) {
			initiator_sockets.push_back( new VisualInitSocket( (*init_iter), cur_mod_id) );
		}
		for(risc::TargetSocketVectorIter targ_iter = m->get_definition()->targetsockets_.begin();
			targ_iter != m->get_definition()->targetsockets_.end();
			targ_iter++){
			target_sockets.push_back( new VisualTargSocket( (*targ_iter), cur_mod_id) );
		}

		
	}
};

VisualModule* initial_walk(risc::Instance* mod) {
	VisualModule* cur_mod = new VisualModule(mod);

	std::list<risc::InstanceTree> temp2 = risc::get_design()->get_root().find_all_instances_of(cur_mod->mod->get_symbol());
	int cur_mod_id = -1;
	for (std::list<risc::InstanceTree>::iterator list_it = temp2.begin();
		list_it != temp2.end();
		list_it++) {
				if(list_it->get_object()->get_name() == cur_mod->mod->get_name()) {	
					cur_mod_id = risc::get_design()->path_instance_mapper_->get_id(*list_it);
					break;
				}
	}
	if(cur_mod_id == -1) {
		printf("could not find current module id!\n");
	}

	for(risc::PrimitiveChannelInstanceVectorIter channel_iter = cur_mod->mod->get_definition()->sub_prim_channel_.begin();
		channel_iter != cur_mod->mod->get_definition()->sub_prim_channel_.end();
		channel_iter++) {
		cur_mod->sub_mod_channel_map[std::make_pair(cur_mod_id,(*channel_iter)->get_symbol())] = -1;
	}
	for(risc::HierarchicalChannelInstanceVectorIter channel_iter = cur_mod->mod->get_definition()->sub_hier_channel_.begin();
		channel_iter != cur_mod->mod->get_definition()->sub_hier_channel_.end();
		channel_iter++) {
		cur_mod->sub_mod_channel_map[std::make_pair(cur_mod_id,(*channel_iter)->get_symbol())] = -1;
	}


	for(risc::ModuleInstanceVectorIter submod_iter = cur_mod->mod->get_definition()->sub_modules_.begin();
		submod_iter != cur_mod->mod->get_definition()->sub_modules_.end();
		submod_iter++) {
		
		cur_mod->sub_modules.push_back(initial_walk(*submod_iter));
		
		for(std::vector<VisualInitSocket*>::iterator init_iter = cur_mod->sub_modules.back()->initiator_sockets.begin();
			init_iter != cur_mod->sub_modules.back()->initiator_sockets.end();
			init_iter++) {

			std::pair<int, SgVariableDefinition*> p1 = {(*init_iter)->instance_id,(*init_iter)->ptr->get_ast_node()};
			risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
		
			if(!init_target_tree.path_.empty()) {
				int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);

				for(std::vector<VisualInitSocket*>::iterator init2_iter = cur_mod->initiator_sockets.begin();
					init2_iter != cur_mod->initiator_sockets.end();
					init2_iter++) {
					if(init_target_tree.get_object_symbol() == (*init2_iter)->ptr->get_symbol()
						&& init_target_id == (*init2_iter)->instance_id ) {
						(*init_iter)->HB = true;
						(*init2_iter)->HB = true;
						cur_mod->HB_init_sockets[std::make_pair(init_target_id,init_target_tree.get_object_symbol())]= init2_iter-cur_mod->initiator_sockets.begin();
					}
				}

				for( int i = 0; i < cur_mod->sub_modules.size() - 1; i++) {
					for(std::vector<VisualTargSocket*>::iterator targ_iter = cur_mod->sub_modules[i]->target_sockets.begin();
						targ_iter != cur_mod->sub_modules[i]->target_sockets.end();
						targ_iter++) {

						if(init_target_tree.get_object_symbol() == (*targ_iter)->ptr->get_symbol()
							&& init_target_id == (*targ_iter)->instance_id) {
							(*init_iter)->HB_connected = true;

							if( (*targ_iter)->HB_connected != true ) {
								(*targ_iter)->HB_connected = true;
								cur_mod->sub_mod_HB_map[std::make_pair((*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol())] = -1;	

							}
						}		
					}
				}
			}
		}
		for(std::vector<VisualTargSocket*>::iterator targ_iter = cur_mod->sub_modules.back()->target_sockets.begin();
			targ_iter != cur_mod->sub_modules.back()->target_sockets.end();
			targ_iter++) {

			for(std::vector<VisualTargSocket*>::iterator targ2_iter = cur_mod->target_sockets.begin();
				targ2_iter != cur_mod->target_sockets.end();
				targ2_iter++) {
				std::pair<int, SgVariableDefinition*> p1 = {(*targ2_iter)->instance_id,(*targ2_iter)->ptr->get_ast_node()};
				risc::InstanceTree targ_target_tree = ordered_tree_mapping[p1];
				if(!targ_target_tree.path_.empty()) {
					int targ_target_id = risc::get_design()->path_instance_mapper_->get_id(targ_target_tree);
					if(targ_target_tree.get_object_symbol() == (*targ_iter)->ptr->get_symbol()
						&& targ_target_id == (*targ_iter)->instance_id) {
						(*targ_iter)->HB = true;
						(*targ2_iter)->HB = true;
						cur_mod->HB_targ_sockets[std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())] = 
							targ2_iter - cur_mod->target_sockets.begin();
					}
				}
			}

			for( int i = 0; i < cur_mod->sub_modules.size() - 1; i++) {
				for(std::vector<VisualInitSocket*>::iterator init_iter = cur_mod->sub_modules[i]->initiator_sockets.begin();
					init_iter != cur_mod->sub_modules[i]->initiator_sockets.end();
					init_iter++) {
					std::pair<int, SgVariableDefinition*> p1 = {(*init_iter)->instance_id,(*init_iter)->ptr->get_ast_node()};
					risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
					
					if(!init_target_tree.path_.empty()) {
						int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);
						if(init_target_tree.get_object_symbol() == (*targ_iter)->ptr->get_symbol()
							&& init_target_id == (*targ_iter)->instance_id) {
							(*init_iter)->HB_connected = true;

							if( (*targ_iter)->HB_connected != true) {
								(*targ_iter)->HB_connected = true;
								cur_mod->sub_mod_HB_map[std::make_pair((*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol())] = -1;
							}
						}
					}
				}
			}
		}
	}
	return cur_mod;	
}

struct VisualCommandArg {
	std::vector<int> int_args;
	std::vector<double> double_args;
	std::string string_arg;
};

void visual_cairo_set_line_width(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_set_line_width(cr,func_args.int_args[0]);
}

void visual_cairo_set_source_rgb(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_set_source_rgb(cr,func_args.double_args[0],func_args.double_args[1],func_args.double_args[2]);
}

void visual_cairo_move_to(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_move_to(cr,func_args.int_args[0],func_args.int_args[1]);
}

void visual_cairo_line_to(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_line_to(cr,func_args.int_args[0],func_args.int_args[1]);
}

void visual_cairo_rectangle(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_rectangle(cr,func_args.int_args[0],func_args.int_args[1],func_args.int_args[2],func_args.int_args[3]);
}

void visual_cairo_set_source_rgba( cairo_t *cr, VisualCommandArg& func_args) {
	cairo_set_source_rgba(cr,func_args.double_args[0],func_args.double_args[1],func_args.double_args[2],func_args.double_args[3]);
}

void visual_cairo_set_line_cap(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_set_line_cap(cr,CAIRO_LINE_CAP_ROUND);
}

void visual_cairo_set_font_size(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_set_font_size(cr,func_args.int_args[0]);
}

void visual_cairo_show_text(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_show_text(cr, func_args.string_arg.c_str());
}

void visual_cairo_curve_to(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_curve_to(cr
		,func_args.int_args[0]
		,func_args.int_args[1]
		,func_args.int_args[2]
		,func_args.int_args[3]
		,func_args.int_args[4]
		,func_args.int_args[5]
		);
}

void visual_cairo_stroke(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_stroke(cr);
}

void visual_cairo_fill(cairo_t *cr, VisualCommandArg& func_args) {
	cairo_fill(cr);
}

/**
 * GLOBAL VARIABLES
 *
 */
static VisualModule* top_;
std::unordered_map<std::string, void (*)(cairo_t*, VisualCommandArg&)> cairo_commands;
std::list< std::pair< std::string, VisualCommandArg > > cairo_cmd_list;
static double global_unscaled_top_width = 0; //needed for pdf export
static double global_unscaled_top_height = 0; //needed for pdf export
static bool global_test_mode = false;
static char* global_input_file_name;

std::unordered_map< std::pair<int,SgSymbol*>, VisualDrawnSocket*,pairhash> global_drawn_sockets;

cairo_surface_t* surface;

/**
 * END GLOBAL VARIABLES
 *
 */

void visual_cairo_slw_save(int width) {
	VisualCommandArg args;
	args.int_args.push_back(width);
	cairo_cmd_list.push_back({"set_line_width",args});
}

void visual_cairo_ssrgb_save(double r, double g, double b) {
	VisualCommandArg args;
	args.double_args.push_back(r);
	args.double_args.push_back(g);
	args.double_args.push_back(b);
	cairo_cmd_list.push_back({"set_source_rgb",args});	
}

void visual_cairo_mt_save(int x, int y) {
	VisualCommandArg args;
	args.int_args.push_back(x);
	args.int_args.push_back(y);
	cairo_cmd_list.push_back({"move_to",args});
}

void visual_cairo_lt_save(int x, int y) {
	VisualCommandArg args;
	args.int_args.push_back(x);
	args.int_args.push_back(y);
	cairo_cmd_list.push_back({"line_to",args});
}

void visual_cairo_r_save(int x, int y, int width, int height) {
	VisualCommandArg args;
	args.int_args.push_back(x);
	args.int_args.push_back(y);
	args.int_args.push_back(width);
	args.int_args.push_back(height);
	cairo_cmd_list.push_back({"rectangle",args});
}

void visual_cairo_f_save() {
	cairo_cmd_list.push_back({"fill",VisualCommandArg()});
}

void visual_cairo_ssrgba_save(double r, double g, double b, double trans) {
	VisualCommandArg args;
	args.double_args.push_back(r);
	args.double_args.push_back(g);
	args.double_args.push_back(b);
	args.double_args.push_back(trans);
	cairo_cmd_list.push_back({"set_source_rgba",args});
}

void visual_cairo_slc_save() {
	cairo_cmd_list.push_back({"set_line_cap",VisualCommandArg()});
}

void visual_cairo_s_save() {
	cairo_cmd_list.push_back({"stroke",VisualCommandArg()});
}

void visual_cairo_sfs_save(int size) {
	VisualCommandArg args;
	args.int_args.push_back(size);
	cairo_cmd_list.push_back({"set_font_size",args});
}

void visual_cairo_st_save(std::string str) {
	VisualCommandArg args;
	args.string_arg = str;
	cairo_cmd_list.push_back({"show_text",args});
}

void visual_cairo_ct_save(int x1, int y1, int x2, int y2, int x3, int y3) {
	VisualCommandArg args;
	args.int_args.push_back(x1);
	args.int_args.push_back(y1);
	args.int_args.push_back(x2);
	args.int_args.push_back(y2);
	args.int_args.push_back(x3);
	args.int_args.push_back(y3);
	cairo_cmd_list.push_back({"curve_to",args});
}



int draw_HB_socket_space(VisualModule* cur_mod, int& ycur) {
	int initial_y = ycur;
	if(cur_mod->sub_mod_HB_map.size() != 0) {
		ycur += D_SPACE;
	}
	for(auto iter = cur_mod->sub_mod_HB_map.begin();
		iter != cur_mod->sub_mod_HB_map.end();
		iter++) {
		//cur_mod->sub_mod_HB_map[iter->first] = ycur;
		iter->second = ycur;
		ycur += PRIM_CHANNEL_DISTANCE;
	}
	return ycur - initial_y;
}

void draw_HB_connect_space(cairo_t* cr,VisualModule* cur_mod, int xcur, int length) {
	for(auto iter = cur_mod->sub_mod_HB_map.begin();
		iter != cur_mod->sub_mod_HB_map.end();
		iter++) {
		visual_cairo_slw_save(2);
		visual_cairo_ssrgb_save(0,0,0);
		visual_cairo_mt_save(xcur+D_SPACE,iter->second);
		visual_cairo_lt_save(xcur+D_SPACE+length,iter->second);
		visual_cairo_s_save();	
	}
}

void check_ports(VisualModule* cur_mod) {
	for(auto mod_iter = cur_mod->sub_modules.begin();
		mod_iter != cur_mod->sub_modules.end();
		mod_iter++) {
		check_ports(*mod_iter);
	}
	std::list<risc::InstanceTree> temp2 = risc::get_design()->get_root().find_all_instances_of(cur_mod->mod->get_symbol());
	int cur_mod_id = -1;
	for (std::list<risc::InstanceTree>::iterator list_it = temp2.begin();
		list_it != temp2.end();
		list_it++) {
				if(list_it->get_object()->get_name() == cur_mod->mod->get_name()) {	
					cur_mod_id = risc::get_design()->path_instance_mapper_->get_id(*list_it);
					break;
				}
	}
	if(cur_mod_id == -1) {
		printf("could not find current module id!\n");
	}

	for(auto inport_iter = cur_mod->mod->get_definition()->inports_.begin();
		inport_iter != cur_mod->mod->get_definition()->inports_.end();
		inport_iter++) {
		std::pair<int, SgVariableDefinition*> p1 = {cur_mod_id,(*inport_iter)->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			
		if(!init_target_tree.path_.empty()) {
			std::cout << "inport " << (*inport_iter)->get_name()  << " is binded to " 
				<< init_target_tree.get_object()->get_name() << "\n";
		}
	
	}

	for(auto inoutport_iter = cur_mod->mod->get_definition()->inoutports_.begin();
		inoutport_iter != cur_mod->mod->get_definition()->inoutports_.end();
		inoutport_iter++) {
		std::pair<int, SgVariableDefinition*> p1 = {cur_mod_id,(*inoutport_iter)->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			
		if(!init_target_tree.path_.empty()) {
			std::cout << "inoutport " << (*inoutport_iter)->get_name()  << " is binded to " 
				<< init_target_tree.get_object()->get_name() << "\n";
		}
	
	}

	for(auto outport_iter = cur_mod->mod->get_definition()->outports_.begin();
		outport_iter != cur_mod->mod->get_definition()->outports_.end();
		outport_iter++) {
		std::pair<int, SgVariableDefinition*> p1 = {cur_mod_id,(*outport_iter)->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			
		if(!init_target_tree.path_.empty()) {
			std::cout << "outport " << (*outport_iter)->get_name()  << " is binded to " 
				<< init_target_tree.get_object()->get_name() << "\n";
		}
	
	}

	for(auto prim_channel_iter = cur_mod->mod->get_definition()->sub_prim_channel_.begin();
		prim_channel_iter != cur_mod->mod->get_definition()->sub_prim_channel_.end();
		prim_channel_iter++) {
		std::pair<int, SgVariableDefinition*> p1 = {cur_mod_id,(*prim_channel_iter)->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			std::cout << "prim channel " << (*prim_channel_iter)->get_name() << "\n";
		if(!init_target_tree.path_.empty()) {
			std::cout << "prim channel " << (*prim_channel_iter)->get_name()  << " is binded to " 
				<< init_target_tree.get_object()->get_name() << "\n";
		}
	}

	for(auto hier_channel_iter = cur_mod->mod->get_definition()->sub_hier_channel_.begin();
		hier_channel_iter != cur_mod->mod->get_definition()->sub_hier_channel_.end();
		hier_channel_iter++) {
		std::pair<int, SgVariableDefinition*> p1 = {cur_mod_id,(*hier_channel_iter)->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
		std::cout << "hier channel " << (*hier_channel_iter)->get_name()  << "\n";
		if(!init_target_tree.path_.empty()) {
			std::cout << "hier channel " << (*hier_channel_iter)->get_name()  << " is binded to " 
				<< init_target_tree.get_object()->get_name() << "\n";
		}

	}

}

void check_HB_sockets(VisualModule* cur_mod) {
	int HB_connected_count = 0;
	for(auto HB_connected_iter = cur_mod->sub_mod_HB_map.begin();
		HB_connected_iter != cur_mod->sub_mod_HB_map.end();
		HB_connected_iter++) {
		HB_connected_count++;
	}
	std::cout << "Module " << cur_mod->mod->get_name() << " has " << HB_connected_count << " amount of HB_connections\n";

	for(std::vector<VisualInitSocket*>::iterator init_iter = cur_mod->initiator_sockets.begin();
		init_iter != cur_mod->initiator_sockets.end();
		init_iter++) {
		if( (*init_iter)->HB_connected == true) {
			std::cout << (*init_iter)->ptr->get_name() << " is a initiator that is HB_connected\n";
		}
		if( (*init_iter)->HB == true) {
			std::cout << (*init_iter)->ptr->get_name() << " is a initiator that is HB\n";
		}

	}
	for(std::vector<VisualTargSocket*>::iterator targ_iter = cur_mod->target_sockets.begin();
		targ_iter != cur_mod->target_sockets.end();
		targ_iter++) {
		if ( (*targ_iter)->HB_connected == true ) {
			std::cout << (*targ_iter)->ptr->get_name() << " is a target that is HB_connected\n";
		}
		if ( (*targ_iter)->HB == true ) {
			std::cout << (*targ_iter)->ptr->get_name() << " is a target that is HB\n";
		}

	}

	for(std::vector<VisualModule*>::iterator mod_iter = cur_mod->sub_modules.begin();
		mod_iter != cur_mod->sub_modules.end();
		mod_iter++) {
		check_HB_sockets((*mod_iter));		
	}

}

struct VisualPort {
	VisualPort(int x, int y, risc::Port* port, int i_id, SgVariableDefinition* symbol_ptr) {
		xcor = x;
		ycor = y;
		port_ptr = port;
		instance_id = i_id;
		var_symbol = symbol_ptr;
	}

	int ycor;
	int xcor;
	risc::Port* port_ptr;

	int instance_id;
	SgVariableDefinition* var_symbol;
};





int format_color_counter(int counter, int max_size) {
	while(counter >= max_size) {
		counter -= max_size;
	}

	return counter;
}

/**
 * \struct	Argument Formatter
 * \brief	This struct handles all commandline parsing
 *
 */
struct ArgumentFormatter {
	int lowest_level;

	bool spec_printing_mod;
	std::string top_mod;

	bool black_white_only;

	double scale_factor;

	bool draw_ports;

	bool pending_update;

	void delete_argv_element(int &cur_index, int &argc, char**argv) {
		for(int i = cur_index; i < argc - 1; i++) {
			argv[i] = argv[i + 1];
		}
		argc = argc - 1;
		cur_index = cur_index - 1;
	}

	ArgumentFormatter(int &argc, char** argv) {
		pending_update = true;
	
		black_white_only = false;
	
		spec_printing_mod = false;	
		top_mod = "";
		
		lowest_level = 0;
		
		scale_factor = 1;

		int delete_counter = 0;
	
		draw_ports = true;
	
		for(int i = 0; i < argc; i++) {
			std::string str(argv[i]);
			
			if(str == "-bw") {
				black_white_only = true;
				delete_argv_element(i,argc,argv);
			}
			else if(str == "-tm") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
					top_mod = std::string(argv[i]);
					delete_argv_element(i,argc,argv);

				}
			}
			else if(str == "-ll") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
				std::string str(argv[i]);
				std::stringstream intConverter;
				intConverter << str;
				intConverter >> lowest_level;
				delete_argv_element(i,argc,argv);
				}
			}
			else if(str == "-s") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
				std::string str(argv[i]);
				std::stringstream doubleConverter;
				doubleConverter << str;
				doubleConverter >> scale_factor;
				delete_argv_element(i,argc,argv);

				}
			}
			else if(str == "-h") {
				print_help_menu();
			}
			else if(str == "-np") {
				delete_argv_element(i,argc,argv);
				draw_ports = false;
			}
		}
	}

	void toggle_ports(){
		draw_ports = !draw_ports;
	}

	bool is_drawing_ports() {
		return draw_ports;
	}

	void print_help_menu(){
		std::cout << "Command line options for visualization tool:" << std::endl;
		std::cout << "-h : Lists command line options and their usages." << std::endl;
		std::cout << "-bw : Modules are drawn without color." << std::endl;
		std::cout << "-tm [module name] : Only draw [module name]." << std::endl;
		std::cout << "-ll [integer] : Draw only a certain depth in the hierarchy given by [integer}." << std::endl;
		std::cout << "-np : Draws module hierarchy without ports or channels." << std::endl;
		std::cout << "-s [floating point number] : Scale the drawing size by [floating point number]. If [floating point number] = 0.5, then the size of the drawing is scaled by 50 percent." << std::endl;
	}

	void increment_depth() {
		if(lowest_level != 0) {
		lowest_level++;
		}
	}

	void decrement_depth() {
		if(lowest_level <= 1) return;
		else {
			lowest_level--;
		}
	}

	void toggle_depth() {
		if(lowest_level != 0) {
			lowest_level = 0;
		}
		else {
			lowest_level = 1;
		}
	}

	int get_depth_var() {
		return lowest_level;
	}

	void increment_scale_factor_by5percent() {
		scale_factor += 0.05;
	}

	void decrement_scale_factor_by5percent() {
		scale_factor -= 0.05;
	}

	void set_scale_factor(double factor) {
		scale_factor = factor;
	}

	double get_scale_factor() {
		return scale_factor;
	}

	bool is_bw_only() {
		return black_white_only;
	}

	void toggle_bw() {
		if(black_white_only) {
			black_white_only = false;
		}
		else {
			black_white_only = true;
		}
	}

	void unset_print_tag(std::string cur_mod) {
		if(spec_printing_mod && cur_mod == top_mod) {
			spec_printing_mod = false;
		}
	}
	
	bool get_set_print_tag_true(std::string cur_mod) {
		if(top_mod != "") {
			if(spec_printing_mod) {
				return true;
			}
			else if(cur_mod == top_mod) {
				spec_printing_mod = true;
				return spec_printing_mod;
			}
			else {
				return false;
			}
		}
		else {
			return true;;
		}
	}

	bool is_this_lowest_level(int cur_level) {
		if(lowest_level <= 0) {
			return false;
		}
		else {
			if(cur_level == lowest_level) {
				return true;
			}
			else {
				return false;
			}
		}
	}

};

/**
 *  \struct	Colorer
 *  \brief	This struct handles the coloring of modules based on their class type
 *
 */
struct Colorer {

	std::vector<risc::Class*> mods;

	int get_color(risc::Instance* mod_param) {
		int i;
		risc::Class* uncolored_mod = mod_param->get_definition();
		for(i = 0; i < mods.size(); i++) {
			if(mods.at(i) == uncolored_mod) {
				return format_color_counter(i,COLOR_PALET_FORMATTER);
			}
		}
		mods.push_back(uncolored_mod);
		return format_color_counter(i,COLOR_PALET_FORMATTER);
		
	}
};

/**
 *  \struct	Box
 *  \brief	This struct represents a module
 *
 */
struct Box {
	
	Box(int x, int y, int w, int h) {
		x_box_cor = x;
		y_box_cor = y;
		box_width = w;
		box_height = h;
	}
	
	int x_box_cor;
	int y_box_cor;
	int box_width;
	int box_height;
};

/**
 * \struct	RGB
 * \brief	This function represents a pixel with its corresponding red, green and blue
 * 		intensities.
 */
struct RGB {
	RGB(double r, double g, double b) {
	r_val = r;
	g_val = g;	
	b_val = b;
	}

	double r_val;
	double g_val;
	double b_val;

};

void draw_color_fill(cairo_t* cr,
	int xcur,
	int ycur,
	int width,
	int height,
	double color_r,
	double color_g,
	double color_b) {
	
	visual_cairo_ssrgba_save(color_r,color_g,color_b,TRANS_VAR);
	visual_cairo_r_save(xcur,ycur,width,height);
	visual_cairo_f_save();

}

int find_largest_char_port(risc::Instance *m) {
	int largest_string = 0;
	for(risc::InPortVectorIter inport_iter = m->get_definition()->inports_.begin();
		inport_iter != m->get_definition()->inports_.end();
		inport_iter++) {
		if( ( (*inport_iter)->get_name()).length() > largest_string ) {
			largest_string = ( (*inport_iter)->get_name()).length();
		}
	}
	for(risc::OutPortVectorIter outport_iter = m->get_definition()->outports_.begin();
		outport_iter != m->get_definition()->outports_.end();
		outport_iter++) {
			if( ((*outport_iter)->get_name()).length() > largest_string ) {
				largest_string = ( (*outport_iter)->get_name()).length();
			}
		}
	for(risc::InOutPortVectorIter inoutport_iter = m->get_definition()->inoutports_.begin();
		inoutport_iter != m->get_definition()->inoutports_.end();
		inoutport_iter++) {
		if( ( (*inoutport_iter)->get_name()).length() > largest_string ) {
			largest_string = ( (*inoutport_iter)->get_name()).length();
		}
	}
	
	return largest_string;
}

void draw_dot(cairo_t* cr, int x, int y) {
	
	visual_cairo_ssrgb_save(0.1,0.1,0.1);
	visual_cairo_slw_save(DOT_WIDTH);
	visual_cairo_slc_save();
	visual_cairo_mt_save(x,y);
	visual_cairo_lt_save(x,y);
	visual_cairo_s_save();
}

void draw_port(cairo_t* cr,int xcur, int ycur,risc::Port* port) {
	
	visual_cairo_ssrgb_save(0.1,0.1,0.1);
	//draw_dot(cr,xcur + max_distance/2,ycur);
	draw_dot(cr,xcur,ycur);	

	//visual_cairo_mt_save(xcur + max_distance/2 - ( port->get_name()).length()/2 * TEXT_CHAR,ycur + 10);
	visual_cairo_mt_save(xcur + PORT_TEXT_XOFFSET,ycur + PORT_TEXT_YOFFSET);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(( port->get_name()).c_str());
}

int draw_base_ports_wrapper(cairo_t* cr,int x, int y, VisualModule *cur_mod, std::vector<VisualPort> &VisualPortVector) {
	int port_distance = find_largest_char_port(cur_mod->mod) * TEXT_CHAR;
	int total_x_offset = 0;
	if(!cur_mod->mod->get_definition()->inports_.empty()
		|| !cur_mod->mod->get_definition()->outports_.empty()
		|| !cur_mod->mod->get_definition()->inoutports_.empty()) {

			total_x_offset += (cur_mod->mod->get_ast_type_name().length()) * TEXT_CHAR + D_SPACE;
			x += (cur_mod->mod->get_ast_type_name().length()) * TEXT_CHAR + D_SPACE;
	}

	for(risc::InPortVectorIter inport_iter = cur_mod->mod->get_definition()->inports_.begin();
		inport_iter != cur_mod->mod->get_definition()->inports_.end();
		inport_iter++) {
		draw_port(cr,x,y,(*inport_iter));
		
		VisualPortVector.push_back(VisualPort(x,y,(*inport_iter),cur_mod->instance_id,(*inport_iter)->get_ast_node()));

		x += port_distance + PORT_TEXT_XOFFSET;
		total_x_offset += port_distance + PORT_TEXT_XOFFSET;


	}
        for(risc::OutPortVectorIter outport_iter = cur_mod->mod->get_definition()->outports_.begin();
                outport_iter != cur_mod->mod->get_definition()->outports_.end();
                outport_iter++) {
                draw_port(cr,x,y,(*outport_iter));

		VisualPortVector.push_back(VisualPort(x,y,(*outport_iter),cur_mod->instance_id,(*outport_iter)->get_ast_node()));

                x += port_distance + PORT_TEXT_XOFFSET;
                total_x_offset += port_distance + PORT_TEXT_XOFFSET;

        }
        for(risc::InOutPortVectorIter inoutport_iter = cur_mod->mod->get_definition()->inoutports_.begin();
                inoutport_iter != cur_mod->mod->get_definition()->inoutports_.end();
                inoutport_iter++) {
                draw_port(cr,x,y,(*inoutport_iter));
                
		VisualPortVector.push_back(VisualPort(x,y,(*inoutport_iter),cur_mod->instance_id,(*inoutport_iter)->get_ast_node()));

		x += port_distance + PORT_TEXT_XOFFSET;
                total_x_offset += port_distance + PORT_TEXT_XOFFSET;
		
        }
	
	return total_x_offset;
}

void draw_connect_line(cairo_t* cr, int x1, int y1, int x2, int y2) {
	visual_cairo_slw_save(1);
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_mt_save(x1,y1);
	visual_cairo_lt_save(x1,y2/2);
	visual_cairo_lt_save(x2,y2/2);
	visual_cairo_lt_save(x2,y2);
	visual_cairo_s_save();
}

void draw_line(cairo_t* cr, int x1, int y1, int x2, int y2) {
	
	static bool collision_manager = false;
	if(cr == NULL) {
		collision_manager = false;
		return;
	}
	else if(cr != NULL 
			&& x1 == 0
			&& x2 == 0
			&& y1 == 0
			&& y2 == 0) {
					collision_manager = true;
					return;
	}


	visual_cairo_slw_save(1);
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_mt_save(x1,y1);

	if(x1 == x2 || y1 == y2) {
		visual_cairo_lt_save(x2,y2);
		collision_manager = false;
	}
	else {
		int y_distance = y2 - y1;
		double scaler;

		if(collision_manager) {
			scaler = 0.75;
			collision_manager = false;
		}
		else {
			collision_manager = true;
			scaler = 0.5;
		}

		visual_cairo_lt_save(x1,y1 + y_distance*scaler);
		
		visual_cairo_mt_save(x1,y1 + y_distance*scaler);
		visual_cairo_lt_save(x2,y1 + y_distance*scaler);
		
		visual_cairo_mt_save(x2,y1 + y_distance*scaler);
		visual_cairo_lt_save(x2,y2);

	}
	visual_cairo_s_save();
}

/*
 * \fn		draw_compared_ports
 * \brief	This function draws the ports on the module.
 */
int draw_compared_ports(cairo_t* cr, int x, int y, std::vector<VisualPort> &lower_ports,std::vector<VisualPort> &current_ports, VisualModule *cur_mod, risc::Instance *lower_mod,int &prev_end,
	std::unordered_map<int,bool>& drawn_inports,
	std::unordered_map<int,bool>& drawn_outports,
	std::unordered_map<int,bool>& drawn_inoutports) {
	int port_distance = find_largest_char_port(cur_mod->mod) * TEXT_CHAR;
	int temp_prev_port_xcor = 0;
	int temp_prev_port_name = 0;
	int temp_adder = 0;
	int max_xcor = 0;
	SgVariableSymbol* temp_low_symbol;

	SgVariableSymbol* temp_high_symbol;
	int temp_high_instance_id;

	SgVariableSymbol* symbol_comparee;
	int temp_comparee_instance_id;
	bool port_found = false;

	//first iterate through child ports
	for(int i = 0; 
		i < lower_ports.size();
		i++) {
			/*			
			temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
			temp_high_symbol = get_mapped_symbol(temp_low_symbol,
					dynamic_cast<risc::Module*>(cur_mod->mod->get_definition()),
					lower_mod->get_ast_node());
			*/
			std::pair<int, SgVariableDefinition*> p1 = {lower_ports[i].instance_id,lower_ports[i].port_ptr->get_ast_node()};
			risc::InstanceTree port_target_tree = ordered_tree_mapping[p1];	
			
			if(!port_target_tree.path_.empty()) {
				temp_high_instance_id = risc::get_design()->path_instance_mapper_->get_id(port_target_tree);
				temp_high_symbol = port_target_tree.get_object_symbol();
			}

			for(risc::InPortVectorIter inport_iter = cur_mod->mod->get_definition()->inports_.begin();
				inport_iter != cur_mod->mod->get_definition()->inports_.end();
				inport_iter++) {
						symbol_comparee = (*inport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
								temp_adder =  temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*inport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*inport_iter)
								,cur_mod->instance_id,(*inport_iter)->get_ast_node()));
							port_found = true;
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*inport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							drawn_inports[inport_iter - cur_mod->mod->get_definition()->inports_.begin()] = true;
							break;
						}
			}

			if(port_found == true) {
				port_found = false;
				continue;
			}

			for(risc::OutPortVectorIter outport_iter = cur_mod->mod->get_definition()->outports_.begin();
				outport_iter != cur_mod->mod->get_definition()->outports_.end();
				outport_iter++) {

						symbol_comparee = (*outport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
								temp_adder = temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*outport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*outport_iter)
								,cur_mod->instance_id,(*outport_iter)->get_ast_node()));
							port_found = true;
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*outport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							drawn_outports[outport_iter-cur_mod->mod->get_definition()->outports_.begin()] = true;
							break;
						}
			}

			if(port_found == true) {
				port_found = false;
				continue;
			}

			for(risc::InOutPortVectorIter inoutport_iter = cur_mod->mod->get_definition()->inoutports_.begin();
				inoutport_iter != cur_mod->mod->get_definition()->inoutports_.end();
				inoutport_iter++) {

						symbol_comparee = (*inoutport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
									temp_adder = temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*inoutport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*inoutport_iter)
								,cur_mod->instance_id,(*inoutport_iter)->get_ast_node()));
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*inoutport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							drawn_inoutports[inoutport_iter-cur_mod->mod->get_definition()->inoutports_.begin()] = true;
							break;
						}
			}
	}

	/*	
	for(risc::InPortVectorIter inport_iter = cur_mod->mod->get_definition()->inports_.begin();
		inport_iter != cur_mod->mod->get_definition()->inports_.end();
		inport_iter++) {
		if(drawn_inports.count(inport_iter-cur_mod->mod->get_definition()->inports_.begin()) == 0) {
			draw_port(cr,prev_end,y,(*inport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*inport_iter),
				cur_mod->instance_id,(*inport_iter)->get_ast_node()));
			int temp_port_name = (*inport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}

		}

	}

	for(risc::OutPortVectorIter outport_iter = cur_mod->mod->get_definition()->outports_.begin();
		outport_iter != cur_mod->mod->get_definition()->outports_.end();
		outport_iter++) {
		if(drawn_outports.count(outport_iter-cur_mod->mod->get_definition()->outports_.begin())== 0) {
			draw_port(cr,prev_end,y,(*outport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*outport_iter)
				,cur_mod->instance_id,(*outport_iter)->get_ast_node()));
			int temp_port_name = (*outport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}

		}
	}

	for(risc::InOutPortVectorIter inoutport_iter = cur_mod->mod->get_definition()->inoutports_.begin();
		inoutport_iter != cur_mod->mod->get_definition()->inoutports_.end();
		inoutport_iter++) {
		if(drawn_inoutports.count(inoutport_iter-cur_mod->mod->get_definition()->inoutports_.begin())==0) {
			draw_port(cr,prev_end,y,(*inoutport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*inoutport_iter)
				,cur_mod->instance_id,(*inoutport_iter)->get_ast_node()));
			int temp_port_name = (*inoutport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}
		}
	}
	*/

	return max_xcor;
}

int draw_leaf_ports(cairo_t* cr, int x, int y, std::vector<VisualPort> &current_ports, VisualModule *cur_mod, int &prev_end,
	std::unordered_map<int,bool>& drawn_inports,
	std::unordered_map<int,bool>& drawn_outports,
	std::unordered_map<int,bool>& drawn_inoutports) {
	int max_xcor = prev_end;
	
	for(risc::InPortVectorIter inport_iter = cur_mod->mod->get_definition()->inports_.begin();
		inport_iter != cur_mod->mod->get_definition()->inports_.end();
		inport_iter++) {
		if(drawn_inports.count(inport_iter-cur_mod->mod->get_definition()->inports_.begin()) == 0) {
			draw_port(cr,prev_end,y,(*inport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*inport_iter),
				cur_mod->instance_id,(*inport_iter)->get_ast_node()));
			int temp_port_name = (*inport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}

		}

	}

	for(risc::OutPortVectorIter outport_iter = cur_mod->mod->get_definition()->outports_.begin();
		outport_iter != cur_mod->mod->get_definition()->outports_.end();
		outport_iter++) {
		if(drawn_outports.count(outport_iter-cur_mod->mod->get_definition()->outports_.begin())== 0) {
			draw_port(cr,prev_end,y,(*outport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*outport_iter)
				,cur_mod->instance_id,(*outport_iter)->get_ast_node()));
			int temp_port_name = (*outport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}

		}
	}

	for(risc::InOutPortVectorIter inoutport_iter = cur_mod->mod->get_definition()->inoutports_.begin();
		inoutport_iter != cur_mod->mod->get_definition()->inoutports_.end();
		inoutport_iter++) {
		if(drawn_inoutports.count(inoutport_iter-cur_mod->mod->get_definition()->inoutports_.begin())==0) {
			draw_port(cr,prev_end,y,(*inoutport_iter));
			current_ports.push_back(VisualPort(prev_end,y,(*inoutport_iter)
				,cur_mod->instance_id,(*inoutport_iter)->get_ast_node()));
			int temp_port_name = (*inoutport_iter)->get_name().length() * TEXT_CHAR;
			prev_end = prev_end + temp_port_name + PORT_TEXT_XOFFSET;
			if(max_xcor < prev_end) {
				max_xcor = prev_end;
			}
		}
	}

	return max_xcor;

}
void draw_box(cairo_t* cr,int x, int y, int width, int height, bool lowest_level,risc::Instance *m) {
	visual_cairo_ssrgb_save(0.1,0.1,0.1);
	visual_cairo_slw_save(2);
	visual_cairo_r_save(x,y,width,height);
	visual_cairo_s_save();

	visual_cairo_mt_save(x,y + 10);
	visual_cairo_sfs_save(10);
	if(!lowest_level) visual_cairo_st_save((m->get_ast_type_name()+ " " + m->get_name()).c_str());
	else {
		visual_cairo_st_save((m->get_ast_type_name().c_str()));
		visual_cairo_mt_save(x + 5,y + 20);
		visual_cairo_st_save((m->get_name().c_str()));
	}
}	

bool initialize_color_palet(std::vector<RGB> &color_palet) {
	color_palet.push_back(RGB(1,0,0));
	color_palet.push_back(RGB(0,1,0));
	color_palet.push_back(RGB(0,0,1));
	color_palet.push_back(RGB(0,1,1));
	color_palet.push_back(RGB(1,0,1));
	color_palet.push_back(RGB(1,1,0));
	color_palet.push_back(RGB(1,1,1));
	color_palet.push_back(RGB(0,0.5,1));
	color_palet.push_back(RGB(1,0,0.5));
	color_palet.push_back(RGB(0.5,1,0));
	color_palet.push_back(RGB(0.1,0,0));
	color_palet.push_back(RGB(0,0.1,0));
	color_palet.push_back(RGB(0,0,0.1));

	return true;
}

void color_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box>* bounds_param,
	std::vector<RGB> &color_palet,
	int color_select) {

	int max_xcor = xcur + max_width;
	
	RGB &cur_color = color_palet.at(color_select);

	if(bounds_param == NULL) {
		draw_color_fill(cr,xcur,ycur,max_width,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
		return;
	}

	std::vector<Box> &bounds = (*bounds_param);

	for(std::vector<Box>::iterator iter = bounds.begin();
		iter != bounds.end();
		iter++) {
			
			draw_color_fill(cr,xcur,ycur,iter->x_box_cor - xcur,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			draw_color_fill(cr,iter->x_box_cor,ycur,iter->box_width,iter->y_box_cor - ycur,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			draw_color_fill(cr,iter->x_box_cor,iter->y_box_cor + iter->box_height,iter->box_width,ycur + max_height - (iter->y_box_cor + iter->box_height),cur_color.r_val,cur_color.g_val,cur_color.b_val);
			xcur += iter->box_width + (iter->x_box_cor - xcur);
			
			if(&(*iter) == &bounds.back()) {
				draw_color_fill(cr,xcur,ycur,max_xcor - xcur,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			}
		
	}
}

void color_level_dependent_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box>* bounds,
	std::vector<RGB> &color_palet,
	int counter) {
	
	counter = format_color_counter(counter,color_palet.size());

	color_box(cr,xcur,ycur,max_width,max_height,bounds,color_palet,counter);

}

void color_module_dependent_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box> *bounds,
	std::vector<RGB> &color_palet,
	Colorer &colorer,
	risc::Instance *mod) {
	
	int selector = colorer.get_color(mod);

	color_box(cr,xcur,ycur,max_width,max_height,bounds,color_palet,selector);

}	

int draw_space_channels(VisualModule *cur_mod,int xcur, int &ycur) {
/*	int total_offset;
	if(!cur_mod->mod->get_definition()->sub_prim_channel_.empty()
		|| !cur_mod->mod->get_definition()->sub_hier_channel_.empty()) {
		total_offset = D_SPACE;
	}
	else {
		return 0;
	}

	for(risc::PrimitiveChannelInstanceVectorIter channel_iter = cur_mod->mod->get_definition()->sub_prim_channel_.begin();
		channel_iter != cur_mod->mod->get_definition()->sub_prim_channel_.end();
		channel_iter++) {
				total_offset += PRIM_CHANNEL_DISTANCE;
	}
	for(risc::HierarchicalChannelInstanceVectorIter channel_iter = cur_mod->mod->get_definition()->sub_hier_channel_.begin();
		channel_iter != cur_mod->mod->get_definition()->sub_hier_channel_.end();
		channel_iter++) {
				total_offset += PRIM_CHANNEL_DISTANCE;
	}
	ycur += total_offset;
	return total_offset;
*/
	int total_offset;
	if(!cur_mod->sub_mod_channel_map.empty()) {
		total_offset = D_SPACE;
		ycur += D_SPACE;
	}
	else {
		return 0;
	}

	for(std::unordered_map<std::pair<int,SgSymbol*>,int,pairhash>::iterator channel_iter = cur_mod->sub_mod_channel_map.begin();
		channel_iter != cur_mod->sub_mod_channel_map.end();
		channel_iter++) {
		channel_iter->second = ycur;
		//cairo_show_text(channel_name):
		total_offset += PRIM_CHANNEL_DISTANCE;
		ycur += PRIM_CHANNEL_DISTANCE;
	}
	return total_offset;

}

void draw_prim_channel(cairo_t* cr,int xcur, int ycur,int width,risc::Instance* prim_channel) {
	
	visual_cairo_ssrgb_save(0.1,0.1,0.1);
	visual_cairo_mt_save(xcur,ycur);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(( prim_channel->get_name()).c_str());

	visual_cairo_ssrgb_save(0.1,0.1,0.1);
	visual_cairo_mt_save(xcur + prim_channel->get_name().length() * TEXT_CHAR,ycur);
	visual_cairo_lt_save(xcur + width - 2 * D_SPACE,ycur);
	visual_cairo_s_save();
}

void draw_channel_connect(cairo_t* cr, std::vector<VisualPort> &lower_ports, risc::Instance *cur_mod, risc::Instance * lower_mod,int temp_ycur_offset, int y) {
	SgVariableSymbol* temp_low_symbol;
	SgVariableSymbol* temp_high_symbol;
	int temp_high_instance_id;

	SgVariableSymbol* symbol_comparee;

	for(risc::HierarchicalChannelInstanceVectorIter channel_iter = cur_mod->get_definition()->sub_hier_channel_.begin();
		channel_iter != cur_mod->get_definition()->sub_hier_channel_.end();
		channel_iter++) {
		symbol_comparee = (*channel_iter)->get_symbol();

		for(int i = 0; i < lower_ports.size(); i++) {
					/*
					temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
					temp_high_symbol = get_mapped_symbol(temp_low_symbol,
						dynamic_cast<risc::Module*>(cur_mod->get_definition()),
						lower_mod->get_ast_node());
					*/
					std::pair<int, SgVariableDefinition*> p1 = {lower_ports[i].instance_id,lower_ports[i].port_ptr->get_ast_node()};
					risc::InstanceTree port_target_tree = ordered_tree_mapping[p1];	
			
					if(!port_target_tree.path_.empty()) {
						temp_high_instance_id = risc::get_design()->path_instance_mapper_->get_id(port_target_tree);
						temp_high_symbol = port_target_tree.get_object_symbol();
	
						if(symbol_comparee == temp_high_symbol) {
							draw_dot(cr,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
							draw_line(cr,lower_ports.at(i).xcor,y + D_SPACE,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
						}
					}
		}
		temp_ycur_offset -= PRIM_CHANNEL_DISTANCE;
	}

	for(risc::PrimitiveChannelInstanceVectorIter channel_iter = cur_mod->get_definition()->sub_prim_channel_.begin();
		channel_iter != cur_mod->get_definition()->sub_prim_channel_.end();
		channel_iter++) {
		symbol_comparee = (*channel_iter)->get_symbol();

		for(int i = 0; i < lower_ports.size(); i++) {
					/*
					temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
					temp_high_symbol = get_mapped_symbol(temp_low_symbol,
						dynamic_cast<risc::Module*>(cur_mod->get_definition()),
						lower_mod->get_ast_node());
					*/
					std::pair<int, SgVariableDefinition*> p1 = {lower_ports[i].instance_id,lower_ports[i].port_ptr->get_ast_node()};
					risc::InstanceTree port_target_tree = ordered_tree_mapping[p1];	
			
					if(!port_target_tree.path_.empty()) {
						temp_high_instance_id = risc::get_design()->path_instance_mapper_->get_id(port_target_tree);
						temp_high_symbol = port_target_tree.get_object_symbol();

						if(symbol_comparee == temp_high_symbol) {
							draw_dot(cr,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
							draw_line(cr,lower_ports.at(i).xcor,y + D_SPACE,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
						}
					}
		}
		temp_ycur_offset -= PRIM_CHANNEL_DISTANCE;
	}
}

/*
 * \fn		draw_channels_wrapper
 * \brief	This function handles the drawing of channels
 */
void draw_channels_wrapper(cairo_t* cr, risc::Instance *m, int x, int y, int mod_width,int temp_ycur_offset, std::vector<VisualPort> &VisualPortVector) {
	if(!m->get_definition()->sub_prim_channel_.empty() 
		|| !m->get_definition()->sub_hier_channel_.empty()) {
		x += D_SPACE;
		y += D_SPACE;
		for(risc::PrimitiveChannelInstanceVectorIter channel_iter = m->get_definition()->sub_prim_channel_.begin();
			channel_iter != m->get_definition()->sub_prim_channel_.end();
			channel_iter++) {
					draw_prim_channel(cr,x,y,mod_width,(*channel_iter));
					y += PRIM_CHANNEL_DISTANCE;
		}

		for(risc::HierarchicalChannelInstanceVectorIter channel_iter = m->get_definition()->sub_hier_channel_.begin();
			channel_iter != m->get_definition()->sub_hier_channel_.end();
			channel_iter++) {
					draw_prim_channel(cr,x,y,mod_width,(*channel_iter));
					y += PRIM_CHANNEL_DISTANCE;
		}
	}
}

int draw_threads(cairo_t *cr, int xcur, int ycur, risc::Instance* m, int &y_offset_threads) {
	risc::Module* mod = (risc::Module*) m->get_definition();
	int x_offset = 0;
	int max_x_offset = 0;
	int thread_draw_count = 0;
	if(!mod->threads_.empty()) {
		y_offset_threads = D_SPACE + TEXT_CHAR;
		x_offset = D_SPACE;
	}
	for(auto thread_iter = mod->threads_.begin();
		thread_iter != mod->threads_.end();
		thread_iter++) {
		
		thread_draw_count++;
		if(thread_draw_count%5 == 0) {
			y_offset_threads += TEXT_CHAR+2*THREAD_SIZE;
			if(max_x_offset < x_offset) {
				max_x_offset = x_offset;
			}
			x_offset = 0;
		}
		int temp_text_x_offset = 0;
		if( ((*thread_iter)->get_name().length() * TEXT_CHAR)/2 - THREAD_SIZE/2 > 0 ) {
			temp_text_x_offset = ((*thread_iter)->get_name().length() * TEXT_CHAR)/2 - THREAD_SIZE/2;
		}
		visual_cairo_ssrgb_save(0.1,0.1,0.1);
		visual_cairo_slw_save(2);
		//visual_cairo_r_save(xcur+x_offset+temp_text_x_offset,ycur + y_offset_threads,THREAD_SIZE,THREAD_SIZE);
		visual_cairo_mt_save(xcur+x_offset+temp_text_x_offset,ycur + y_offset_threads);
		visual_cairo_lt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE,ycur + y_offset_threads);
		visual_cairo_mt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur + y_offset_threads);
		visual_cairo_ct_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur + y_offset_threads,
										 xcur+x_offset+temp_text_x_offset+THREAD_SIZE/4 , ycur+y_offset_threads+THREAD_SIZE/4,
										 xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2 , ycur+y_offset_threads+THREAD_SIZE/2);
		visual_cairo_mt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur + y_offset_threads+THREAD_SIZE/2);
		visual_cairo_ct_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur + y_offset_threads+THREAD_SIZE/2,
										 xcur+x_offset+temp_text_x_offset+3*THREAD_SIZE/4 , ycur+y_offset_threads+3*THREAD_SIZE/4,
										 xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2 , ycur+y_offset_threads+THREAD_SIZE);
		visual_cairo_mt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE,ycur+y_offset_threads+3*THREAD_SIZE/4);
		visual_cairo_lt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur+y_offset_threads+THREAD_SIZE);
		visual_cairo_mt_save(xcur+x_offset+temp_text_x_offset,ycur+y_offset_threads+3*THREAD_SIZE/4);
		visual_cairo_lt_save(xcur+x_offset+temp_text_x_offset+THREAD_SIZE/2,ycur+y_offset_threads+THREAD_SIZE);

		visual_cairo_s_save();

		visual_cairo_mt_save(xcur+x_offset,ycur+y_offset_threads-TEXT_CHAR);
		visual_cairo_sfs_save(10);
		visual_cairo_st_save(( (*thread_iter)->get_name() ).c_str());

		x_offset+=THREAD_SIZE + (*thread_iter)->get_name().length() * TEXT_CHAR;
	}
	if(x_offset != 0) {
		y_offset_threads += THREAD_SIZE + D_SPACE;
	}
	if(max_x_offset < x_offset || thread_draw_count < 5) {
		return x_offset;
	}
	else {
		return max_x_offset;
	}
}

int draw_HB_init_socket(cairo_t* cr,int xcur, int ycur,std::string name) {
	int x_offset = SOCKET_SIZE + name.length() * TEXT_CHAR;

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_r_save(xcur,ycur-SOCKET_SIZE/2,SOCKET_SIZE,SOCKET_SIZE);
	visual_cairo_s_save();
	visual_cairo_mt_save(xcur+SOCKET_SIZE,ycur+TEXT_CHAR);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(name.c_str());
	
	visual_cairo_slw_save(2);
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_mt_save(xcur,ycur+SOCKET_SIZE/2);
	visual_cairo_lt_save(xcur+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2);
	visual_cairo_mt_save(xcur+SOCKET_SIZE,ycur+SOCKET_SIZE/2);
	visual_cairo_lt_save(xcur+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2);
	visual_cairo_s_save();

	return x_offset;
}

int draw_HB_targ_socket(cairo_t* cr,int xcur, int ycur,std::string name) {
	int x_offset = SOCKET_SIZE + name.length() * TEXT_CHAR;

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_r_save(xcur,ycur-SOCKET_SIZE/2,SOCKET_SIZE,SOCKET_SIZE);
	visual_cairo_s_save();
	visual_cairo_mt_save(xcur+SOCKET_SIZE,ycur+TEXT_CHAR);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(name.c_str());
	
	visual_cairo_slw_save(2);
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_mt_save(xcur,ycur-SOCKET_SIZE/2);
	visual_cairo_lt_save(xcur+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2);
	visual_cairo_mt_save(xcur+SOCKET_SIZE,ycur-SOCKET_SIZE/2);
	visual_cairo_lt_save(xcur+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2);
	visual_cairo_s_save();

	return x_offset;

}

void draw_HB_top_connection(cairo_t* cr,int xcur, int ycur,int ydest) {

	draw_dot(cr, xcur+SOCKET_SIZE/2, ydest);

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_mt_save(xcur+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2);
	visual_cairo_lt_save(xcur+SOCKET_SIZE/2,ydest);
	visual_cairo_s_save();
}

void draw_HB_connection(cairo_t* cr, int xcor_parent, int ycor_parent, VisualDrawnSocket* dest_child) {
	
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_mt_save(dest_child->x,dest_child->y);
	visual_cairo_lt_save(dest_child->x,dest_child->y - D_SPACE/4);
	visual_cairo_mt_save(dest_child->x,dest_child->y - D_SPACE/4);
	visual_cairo_lt_save(xcor_parent,dest_child->y - D_SPACE/4);
	visual_cairo_mt_save(xcor_parent,dest_child->y-D_SPACE/4);
	visual_cairo_lt_save(xcor_parent,ycor_parent);
	visual_cairo_s_save();
}

void draw_general_connection_curve(cairo_t* cr, int xcor, int ycor, int xcor2, int ycor2) {
	int y_offset = abs(xcor-xcor2);

	visual_cairo_slw_save(2);
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_mt_save(xcor,ycor);
	visual_cairo_ct_save(xcor,ycor,
		xcor+(xcor2-xcor)/2 , ycor+(ycor2-ycor)/2 + y_offset,
		xcor2,ycor2);
	visual_cairo_s_save();
}

void draw_init_socket(cairo_t* cr,int xcur, int ycur,std::string name) {
	int x_offset = name.length() * TEXT_CHAR + SOCKET_SIZE;

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_r_save(xcur-SOCKET_SIZE,ycur,SOCKET_SIZE,SOCKET_SIZE);
	visual_cairo_s_save();
	visual_cairo_mt_save(xcur-x_offset,ycur+SOCKET_SIZE/2);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(name.c_str());

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_mt_save(xcur-SOCKET_SIZE,ycur);
	visual_cairo_lt_save(xcur,ycur+SOCKET_SIZE/2);
	visual_cairo_mt_save(xcur-SOCKET_SIZE,ycur+SOCKET_SIZE);
	visual_cairo_lt_save(xcur,ycur+SOCKET_SIZE/2);

	visual_cairo_s_save();
}

int draw_targ_socket(cairo_t* cr,int xcur, int ycur,std::string name) {
	int x_offset = SOCKET_SIZE + name.length() * TEXT_CHAR;

	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_r_save(xcur,ycur,SOCKET_SIZE,SOCKET_SIZE);
	visual_cairo_s_save();
	visual_cairo_mt_save(xcur+SOCKET_SIZE,ycur+SOCKET_SIZE/2);
	visual_cairo_sfs_save(10);
	visual_cairo_st_save(name.c_str());
	
	visual_cairo_ssrgb_save(0,0,0);
	visual_cairo_slw_save(2);
	visual_cairo_mt_save(xcur,ycur);
	visual_cairo_lt_save(xcur+SOCKET_SIZE,ycur+SOCKET_SIZE/2);
	visual_cairo_mt_save(xcur,ycur+SOCKET_SIZE);
	visual_cairo_lt_save(xcur+SOCKET_SIZE,ycur+SOCKET_SIZE/2);
	visual_cairo_s_save();
	return x_offset;

}

int draw_sockets2(cairo_t* cr, VisualModule* cur_mod,std::vector<VisualDrawnSocket*>& submod_drawn_HB_sockets, std::vector<VisualDrawnSocket*>& recent_drawn_HB_sockets,std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash>& parent_mod_HB_map
	,int xcur
	,int ycur
	,int mod_width
	,int x_offset
	,int& y_offset) {
	
	std::unordered_map<int,bool> recent_drawn_HBinit_checker;
	std::unordered_map<int,bool> recent_drawn_HBtarg_checker;

/*
	int init_y_offset = 0;
	int targ_y_offset = 0;
	if(!m->get_definition()->initiatorsockets_.empty() 
		|| !m->get_definition()->targetsockets_.empty()) {
		init_y_offset += D_SPACE;
		targ_y_offset += D_SPACE;
	}
*/
	//if(x_offset < (cur_mod->mod->get_definition()->get_name().size() + cur_mod->mod->get_name().size()) * TEXT_CHAR)
	//	x_offset = (cur_mod->mod->get_name().size() + cur_mod->mod->get_definition()->get_name().size()) * TEXT_CHAR + D_SPACE;

	int nonHB_x_offset = 0;
	int nonHB_targ_y_offset = y_offset;
	int nonHB_init_y_offset = y_offset;

	if(cur_mod->sub_modules.size() == 0 && y_offset == 0) {
		nonHB_targ_y_offset += 2*TEXT_CHAR;
		nonHB_init_y_offset += 2*TEXT_CHAR;
	}		

	for(std::vector<VisualDrawnSocket*>::iterator sub_socket_iter = submod_drawn_HB_sockets.begin();
		sub_socket_iter != submod_drawn_HB_sockets.end();
		sub_socket_iter++) {
		if(cur_mod->HB_init_sockets.count(
			std::make_pair((*sub_socket_iter)->target_instance_id,
				(*sub_socket_iter)->target)) > 0) {

			int index = cur_mod->HB_init_sockets[std::make_pair(
				(*sub_socket_iter)->target_instance_id,(*sub_socket_iter)->target)];

			std::pair<int, SgVariableDefinition*> p1 = {cur_mod->initiator_sockets[index]->instance_id,cur_mod->initiator_sockets[index]->ptr->get_ast_node()};
			risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);
			if((*sub_socket_iter)->x > xcur+x_offset) {
				x_offset = (*sub_socket_iter)->x - xcur - SOCKET_SIZE/2;
			}		
			recent_drawn_HB_sockets.push_back( new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2
				,init_target_tree.get_object_symbol(),init_target_id,true) );
			recent_drawn_HBinit_checker[index] = true;

			if(cur_mod->initiator_sockets[index]->HB_connected ) {
				std::pair<int, SgSymbol*> pk = {init_target_id,init_target_tree.get_object_symbol()};
				draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
			}

			draw_HB_connection(cr,xcur+x_offset+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2,(*sub_socket_iter));
			x_offset += draw_HB_init_socket(cr,xcur+x_offset,ycur,cur_mod->initiator_sockets[index]->ptr->get_name());	
		}
		else if(cur_mod->HB_targ_sockets.count(
			std::make_pair((*sub_socket_iter)->target_instance_id,
				(*sub_socket_iter)->target)) > 0) {

			int index = cur_mod->HB_targ_sockets[std::make_pair(
				(*sub_socket_iter)->target_instance_id, (*sub_socket_iter)->target)];
			if((*sub_socket_iter)->x > xcur+x_offset) {
				x_offset = (*sub_socket_iter)->x - xcur - SOCKET_SIZE/2;
			}
			recent_drawn_HB_sockets.push_back( new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2
				,cur_mod->target_sockets[index]->ptr->get_symbol(),cur_mod->target_sockets[index]->instance_id,false) );
			recent_drawn_HBtarg_checker[index] = true;
			
			if(cur_mod->target_sockets[index]->HB_connected) {
				std::pair<int, SgSymbol*> pk = {cur_mod->target_sockets[index]->instance_id,cur_mod->target_sockets[index]->ptr->get_symbol()};
				draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
			}

			draw_HB_connection(cr,xcur+x_offset+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2,(*sub_socket_iter));
			x_offset += draw_HB_targ_socket(cr,xcur+x_offset,ycur,cur_mod->target_sockets[index]->ptr->get_name());
		}
		else {
			global_drawn_sockets[std::make_pair((*sub_socket_iter)->target_instance_id,(*sub_socket_iter)->target)] = (*sub_socket_iter);;
		}
	}

	for(std::vector<VisualTargSocket*>::iterator targ_iter = cur_mod->target_sockets.begin();
		targ_iter != cur_mod->target_sockets.end();
		targ_iter++) {
		if(!(*targ_iter)->HB) {
			int this_drawn_xcur;
			int this_drawn_ycur;
			if((*targ_iter)->HB_connected) {
				//at this point we know the targ_iter is a top level target HB socket, so we simply draw the HB connection to its init
				std::pair<int, SgSymbol*> pk = {(*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol()};
				draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
				this_drawn_xcur = xcur+x_offset;
				this_drawn_ycur = ycur-SOCKET_SIZE/2;
				x_offset += draw_HB_targ_socket(cr, xcur+x_offset,ycur,(*targ_iter)->ptr->get_name());
			}
			else {
				nonHB_targ_y_offset += 2*SOCKET_SIZE;
				
				std::pair<int,SgVariableDefinition*> p1={(*targ_iter)->instance_id,(*targ_iter)->ptr->get_ast_node()};
				risc::InstanceTree targ_target_tree = ordered_tree_mapping[p1];
				this_drawn_xcur = xcur;
				this_drawn_ycur = ycur+nonHB_targ_y_offset-SOCKET_SIZE/2;

				if(!targ_target_tree.path_.empty()) {
					int targ_target_id = risc::get_design()->path_instance_mapper_->get_id(targ_target_tree);
					if(global_drawn_sockets.count(std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())) > 0
						&& !global_drawn_sockets[std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())]->initiator) {
						//target of targ_iter is already drawn, so we draw the connection
						std::pair<int,SgSymbol*> targ_key = {targ_target_id,targ_target_tree.get_object_symbol()};
						draw_general_connection_curve(cr,this_drawn_xcur,this_drawn_ycur,
							global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);
					}
					else {
						//target of targ_iter is not drawn yet, so we put the notion in global drawn sockets
						global_drawn_sockets[std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())] = new VisualDrawnSocket(this_drawn_xcur,this_drawn_ycur,
															targ_target_tree.get_object_symbol(),
															targ_target_id,true);
					}
				}
				int temp_targ_x_offset = draw_targ_socket(cr,xcur,ycur+nonHB_targ_y_offset-SOCKET_SIZE,(*targ_iter)->ptr->get_name());
				nonHB_targ_y_offset += TEXT_CHAR;
				if(temp_targ_x_offset > nonHB_x_offset) {
					nonHB_x_offset = temp_targ_x_offset;
				}

			}
			//FOR MOST GENERAL BINDING:
			if(global_drawn_sockets.count(std::make_pair((*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol()) ) > 0
				&& global_drawn_sockets[std::make_pair((*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol())]->initiator) {
				//an initiator is binded to the init_iter, so we draw the connection
				std::pair<int,SgSymbol*> targ_key = {(*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol()};
				draw_general_connection_curve(cr,this_drawn_xcur,this_drawn_ycur,
					global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);

			}
			else {
				//the initiator at this point does not havea parent init bounded to it or the parent init has not been drawn
				//we don't know this for sure yet... so we place this initiator as a target in the global drawn sockets
				global_drawn_sockets[ std::make_pair((*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol()) ] = 
					new VisualDrawnSocket(this_drawn_xcur,this_drawn_ycur,(*targ_iter)->ptr->get_symbol(),(*targ_iter)->instance_id,false);
			}
		}
		else if(recent_drawn_HBtarg_checker.count(targ_iter - cur_mod->target_sockets.begin()) == 0) {
			//at this point we know that this is a bottom level HB target socket since no child module's targ socket is binded to it
			//although it is at the bottom level of the HB, it can still be an initiator bound to a target that violates HB criteria
			recent_drawn_HB_sockets.push_back( new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2,
				(*targ_iter)->ptr->get_symbol(),
				(*targ_iter)->instance_id,false ) );

			if((*targ_iter)->HB_connected) {
				std::pair<int, SgSymbol*> pk = {(*targ_iter)->instance_id,(*targ_iter)->ptr->get_symbol()};
				draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
			}
			
			std::pair<int,SgVariableDefinition*> p1={(*targ_iter)->instance_id,(*targ_iter)->ptr->get_ast_node()};
			risc::InstanceTree targ_target_tree = ordered_tree_mapping[p1];
			if(!targ_target_tree.path_.empty()) {
				int targ_target_id = risc::get_design()->path_instance_mapper_->get_id(targ_target_tree);
				if(global_drawn_sockets.count(std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())) > 0
					&& !global_drawn_sockets[std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())]->initiator) {
					//target of targ_iter is already drawn, so we draw the connection
					std::pair<int,SgSymbol*> targ_key = {targ_target_id,targ_target_tree.get_object_symbol()};
					draw_general_connection_curve(cr,xcur+x_offset+SOCKET_SIZE/2,ycur,
						global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);

				}
				else {
					//target of targ_iter is not drawn yet, so we put the notion in global drawn sockets
					global_drawn_sockets[std::make_pair(targ_target_id,targ_target_tree.get_object_symbol())] = new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2,
															targ_target_tree.get_object_symbol(),
															targ_target_id,true);
				}
			}

			x_offset += draw_HB_targ_socket(cr,xcur+x_offset,ycur,(*targ_iter)->ptr->get_name());

		}

	}


	std::vector<VisualInitSocket*> ordinary_inits;
	int nonHB_init_x_offset = 0;
	for(std::vector<VisualInitSocket*>::iterator init_iter = cur_mod->initiator_sockets.begin();
		init_iter != cur_mod->initiator_sockets.end();
		init_iter++) {
		if((*init_iter)->HB && recent_drawn_HBinit_checker.count(init_iter - cur_mod->initiator_sockets.begin()) == 0) {
			//At this point, we should know that the init_iter is at the bottom level of a hierarchical binding
 			std::pair<int, SgVariableDefinition*> p1 = {(*init_iter)->instance_id,(*init_iter)->ptr->get_ast_node()};
			risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
			
			if(!init_target_tree.path_.empty()) {
				int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);

				recent_drawn_HB_sockets.push_back(new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur-SOCKET_SIZE/2,
									init_target_tree.get_object_symbol(),
									init_target_id,true));			
				if((*init_iter)->HB_connected) {
					std::pair<int, SgSymbol*> pk = {init_target_id,init_target_tree.get_object_symbol()};
					draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
				}

			}

			x_offset += draw_HB_init_socket(cr,xcur+x_offset,ycur,(*init_iter)->ptr->get_name());
		}
		else if(!(*init_iter)->HB){
			if((*init_iter)->HB_connected) {
				//at this point we know that the init_iter is a top level initiator, so we simply draw the HB connection to the target
				//that is cached
				std::pair<int, SgVariableDefinition*> p1 = {(*init_iter)->instance_id,(*init_iter)->ptr->get_ast_node()};
				risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];
				if(!init_target_tree.path_.empty()) {
					int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);

					std::pair<int, SgSymbol*> pk = {init_target_id,init_target_tree.get_object_symbol()};
					draw_HB_top_connection(cr,xcur+x_offset,ycur,parent_mod_HB_map.at(pk));
				}

				//FOR MOST GENERAL BINDING:
				if(global_drawn_sockets.count(std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()) ) > 0
					&& global_drawn_sockets[std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol())]->initiator) {
					//an initiator is binded to the init_iter, so we draw the connection
					std::pair<int,SgSymbol*> targ_key = {(*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()};
					draw_general_connection_curve(cr,xcur+x_offset+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2,
						global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);

				}
				else {
					//the initiator at this point does not havea parent init bounded to it or the parent init has not been drawn
					//we don't know this for sure yet... so we place this initiator as a target in the global drawn sockets
					global_drawn_sockets[ std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()) ] = 
						new VisualDrawnSocket(xcur+x_offset+SOCKET_SIZE/2,ycur+SOCKET_SIZE/2,(*init_iter)->ptr->get_symbol(),(*init_iter)->instance_id,false);
				}

				x_offset += draw_HB_init_socket(cr,xcur+x_offset,ycur,(*init_iter)->ptr->get_name());
			}
			else {
				//at this point we don't know if there's more HB sockets which would increase x_offset so we must wait to draw
				//the non-HB init sockets
				ordinary_inits.push_back(*init_iter);
				if((*init_iter)->ptr->get_name().length()*TEXT_CHAR + SOCKET_SIZE > nonHB_init_x_offset) {
					nonHB_init_x_offset = (*init_iter)->ptr->get_name().length()*TEXT_CHAR + SOCKET_SIZE;
				}
				continue;
			}
		}

	}
	
	nonHB_x_offset += nonHB_init_x_offset;
	if(nonHB_x_offset > x_offset) {
		x_offset = nonHB_x_offset;
	}
	if(mod_width > x_offset) {
		x_offset = mod_width;
	}
	for(std::vector<VisualInitSocket*>::iterator init_iter = ordinary_inits.begin();
		init_iter != ordinary_inits.end();
		init_iter++) {
		
		nonHB_init_y_offset += 2*SOCKET_SIZE;
		std::pair<int, SgVariableDefinition*> p1 = {(*init_iter)->instance_id,(*init_iter)->ptr->get_ast_node()};
		risc::InstanceTree init_target_tree = ordered_tree_mapping[p1];	
		
		if(!init_target_tree.path_.empty()) {
			int init_target_id = risc::get_design()->path_instance_mapper_->get_id(init_target_tree);
			if(global_drawn_sockets.count( std::make_pair(init_target_id,init_target_tree.get_object_symbol()) ) > 0
				&& !global_drawn_sockets[std::make_pair(init_target_id,init_target_tree.get_object_symbol())]->initiator ) {
				//the target of the init_iter is alrdy drawn, so we draw the connection between init_iter and its target
				std::pair<int,SgSymbol*> targ_key = {init_target_id,init_target_tree.get_object_symbol()};
				draw_general_connection_curve(cr,xcur+x_offset,ycur+nonHB_init_y_offset-SOCKET_SIZE/2,
					global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);

			}
			else {
				//the target of the init_iter is not drawn yet, so we place a VisualDrawnSocket as a initiator in global drawn
				global_drawn_sockets[std::make_pair(init_target_id,init_target_tree.get_object_symbol())] = new VisualDrawnSocket(
														xcur+x_offset,ycur+nonHB_init_y_offset-SOCKET_SIZE/2,
														init_target_tree.get_object_symbol(),
														init_target_id,true);
			}
		}
		
		//FOR MOST GENERAL BINDING:
		if(global_drawn_sockets.count(std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()) ) > 0
			&& global_drawn_sockets[std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol())]->initiator) {
			//an initiator is binded to the init_iter, so we draw the connection
			std::pair<int,SgSymbol*> targ_key = {(*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()};
			draw_general_connection_curve(cr,xcur+x_offset-SOCKET_SIZE/2,ycur+nonHB_init_y_offset-SOCKET_SIZE/2,
				global_drawn_sockets[targ_key]->x,global_drawn_sockets[targ_key]->y);

		}
		else {
			//the initiator at this point does not havea parent init bounded to it or the parent init has not been drawn
			//we don't know this for sure yet... so we place this initiator as a target in the global drawn sockets
			global_drawn_sockets[ std::make_pair((*init_iter)->instance_id,(*init_iter)->ptr->get_symbol()) ] = 
				new VisualDrawnSocket(xcur+x_offset-SOCKET_SIZE/2,ycur+nonHB_init_y_offset-SOCKET_SIZE/2,(*init_iter)->ptr->get_symbol(),(*init_iter)->instance_id,false);
		}

		draw_init_socket(cr,xcur+x_offset,ycur+nonHB_init_y_offset-SOCKET_SIZE,(*init_iter)->ptr->get_name());
		nonHB_init_y_offset += TEXT_CHAR;
	}

	if(nonHB_targ_y_offset > nonHB_init_y_offset) {
		y_offset = nonHB_targ_y_offset;
	}
	else {
		y_offset = nonHB_init_y_offset;
	}

	return x_offset;

}



/**
 * \fn		draw_recursor
 * \brief	This function is the recursive function that handles the
 * 		control flow of the entire drawing
 */
int draw_recursor(VisualModule *cur_mod, int &param_deepest_level, int &height_tracker, std::vector<VisualPort> &VisualPortVector, risc::MappingInfo& port_cache_mapping_, cairo_t* cr, ArgumentFormatter &args_format,std::vector<VisualDrawnSocket*>& recent_drawn_HB_sockets,std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash>& parent_mod_HB_map) {

	static int cur_xcor = 0;
	static int cur_ycor = 0;
	static int cur_level = 0;
	static std::vector<RGB> color_palet;
	static Colorer colorer;

	static bool initializer = initialize_color_palet(color_palet);
	
	
	if(args_format.get_set_print_tag_true(cur_mod->mod->get_name())) {
		cur_xcor += D_SPACE;
		cur_ycor += D_SPACE;
		cur_level++;

		if(param_deepest_level < cur_level) {
			param_deepest_level = cur_level;
		}
	}
	if(cur_mod->mod->get_definition()->sub_modules_.empty()
		&& cur_mod->mod->get_definition()->sub_hier_channel_.empty()
		&& cur_mod->mod->get_definition()->sub_prim_channel_.empty()
		&& args_format.get_set_print_tag_true(cur_mod->mod->get_name())
		|| (args_format.is_this_lowest_level(cur_level)
		&& args_format.get_set_print_tag_true(cur_mod->mod->get_name())) ) {
		int base_box_width = 0;
		int base_box_height = 0;
		RGB cur_color = color_palet.at(format_color_counter(cur_level,color_palet.size()));
	
		int name_size = 0;
		if( cur_mod->mod->get_ast_type_name().length() >  cur_mod->mod->get_name().length() ) {
			name_size = base_box_width = cur_mod->mod->get_ast_type_name().length() * TEXT_CHAR;
		}
		else {
			name_size = base_box_width = cur_mod->mod->get_name().length() * TEXT_CHAR;
		}

		//drawing threads
		int x_offset_threads = 0;
		int y_offset_threads = 0;
		x_offset_threads = draw_threads(cr,cur_xcor,cur_ycor+2*TEXT_CHAR,cur_mod->mod,y_offset_threads);
		if(x_offset_threads > base_box_width) {
			base_box_width = x_offset_threads;
		}
		base_box_height = y_offset_threads;

		if(args_format.is_drawing_ports()) {
			int x_offset_ports = 0;
			x_offset_ports = draw_base_ports_wrapper(cr,cur_xcor,cur_ycor,cur_mod,VisualPortVector);
			if(x_offset_ports > base_box_width) {
				base_box_width = x_offset_ports;
			}

			int x_offset_sockets = 0;
			int y_offset_sockets = y_offset_threads;

			std::vector<VisualDrawnSocket*> submod_drawn_HB_sockets;
			x_offset_sockets = draw_sockets2(cr,cur_mod,submod_drawn_HB_sockets,recent_drawn_HB_sockets,parent_mod_HB_map,cur_xcor,cur_ycor,base_box_width,x_offset_ports? x_offset_ports : name_size,y_offset_sockets);
			if(x_offset_sockets > base_box_width) {
				base_box_width = x_offset_sockets;
			}
			base_box_height = y_offset_sockets;
		}

		if(base_box_height < MINH) {
			base_box_height = MINH;
		}
				
		draw_box(cr,cur_xcor,cur_ycor,base_box_width,base_box_height,true,cur_mod->mod);

		if(!args_format.is_bw_only()) color_module_dependent_box(cr,cur_xcor,cur_ycor,base_box_width,base_box_height,NULL,color_palet,colorer,cur_mod->mod);


		cur_xcor -= D_SPACE;
		cur_ycor -= D_SPACE;
		cur_level--;
		//height_tracker = MINH;
		height_tracker = base_box_height;
		args_format.unset_print_tag(cur_mod->mod->get_name());
		return base_box_width;
	}
	else {
		int this_box_width = 0;
		int this_box_height = 0;
		int this_box_deepest_level = cur_level;
		int temp_width_adder;
		int temp_height_tracker;
		int temp_xcur_offset = 0;
		int temp_ycur_offset = 0;
		std::vector<Box> box_draw_bounds;
		//std::vector<VisualPort> this_box_VisualPortVector;
		int max_port = 0;
		int temp_port_max;		

		std::unordered_map<int,bool> drawn_inports;
		std::unordered_map<int,bool> drawn_outports;
		std::unordered_map<int,bool> drawn_inoutports;


		int name_width = (cur_mod->mod->get_ast_type_name() + " " +  cur_mod->mod->get_name()).length() * TEXT_CHAR;		
		int end_of_prev_word = name_width + cur_xcor;

		if(args_format.get_set_print_tag_true(cur_mod->mod->get_name())) {
			if(args_format.is_drawing_ports()) {
			temp_ycur_offset = draw_space_channels(cur_mod,cur_xcor,cur_ycor);
			temp_ycur_offset += draw_HB_socket_space(cur_mod,cur_ycor);
			}

			this_box_width += 2 * D_SPACE;
			this_box_width += (cur_mod->mod->get_definition()->sub_modules_.size() - 1) * D_SPACE;
		}
		
		std::vector<VisualDrawnSocket*> submod_drawn_HB_sockets;
		for(std::vector<VisualModule*>::iterator iter = cur_mod->sub_modules.begin();
			iter != cur_mod->sub_modules.end();
			iter++) {

				std::vector<VisualPort> temp_binded_ports;

				temp_width_adder = draw_recursor((*iter),this_box_deepest_level,temp_height_tracker,temp_binded_ports,port_cache_mapping_,cr,args_format,submod_drawn_HB_sockets,cur_mod->sub_mod_HB_map);

				this_box_width += temp_width_adder;
				
				if(temp_height_tracker > this_box_height) {
					this_box_height = temp_height_tracker;
				}

				if(args_format.get_set_print_tag_true(cur_mod->mod->get_name())) {

					if(args_format.is_drawing_ports()) {		
						if(box_draw_bounds.empty() && (!cur_mod->mod->get_definition()->sub_hier_channel_.empty() || !cur_mod->mod->get_definition()->sub_prim_channel_.empty())) {
							draw_line(cr,0,0,0,0);
						}
						else {
							draw_line(NULL,0,0,0,0);
						}
						temp_port_max =  draw_compared_ports(cr,cur_xcor,cur_ycor - temp_ycur_offset,temp_binded_ports,VisualPortVector,cur_mod,(*iter)->mod,end_of_prev_word,
						drawn_inports,
						drawn_outports,
						drawn_inoutports);
						if(max_port < temp_port_max) {
							max_port = temp_port_max;
						}				
						draw_channel_connect(cr, temp_binded_ports, cur_mod->mod, (*iter)->mod,temp_ycur_offset, cur_ycor);
					}

					box_draw_bounds.push_back(Box(D_SPACE + cur_xcor,D_SPACE + cur_ycor,temp_width_adder,temp_height_tracker));
				
					if((*iter)->mod != cur_mod->mod->get_definition()->sub_modules_.back()) {
						temp_xcur_offset += D_SPACE + temp_width_adder;
						cur_xcor += D_SPACE + temp_width_adder;
					}
				}

		}
		if(args_format.is_drawing_ports()) {
			temp_port_max =  draw_leaf_ports(cr,cur_xcor,cur_ycor - temp_ycur_offset,VisualPortVector,cur_mod,end_of_prev_word,
				drawn_inports,
				drawn_outports,
				drawn_inoutports);
			if(max_port < temp_port_max) {
				max_port = temp_port_max;
			}				
		}
		if(args_format.get_set_print_tag_true(cur_mod->mod->get_name())) {
		cur_xcor -= temp_xcur_offset;
		cur_ycor -= temp_ycur_offset;

		if(this_box_width < name_width) {
			this_box_width = name_width;
		}
		if(this_box_width < max_port - cur_xcor) {
			this_box_width = max_port - cur_xcor;
		}
	

		this_box_height = this_box_height + 2 * D_SPACE + temp_ycur_offset;

		//draw threads
		int threads_x_offset = 0;
		int threads_y_offset = 0;
		threads_x_offset = draw_threads(cr,cur_xcor, cur_ycor+this_box_height, cur_mod->mod, threads_y_offset);
		if(threads_x_offset > this_box_width) {
			this_box_width = threads_x_offset;
		}
		this_box_height += threads_y_offset;

		//draw socket!
		if(args_format.is_drawing_ports()) {
			int socket_y_offset = 0;
			int socket_x_offset = 0;

			socket_x_offset = draw_sockets2(cr,cur_mod,submod_drawn_HB_sockets,recent_drawn_HB_sockets,parent_mod_HB_map,cur_xcor,cur_ycor,this_box_width,max_port?max_port-cur_xcor : name_width,socket_y_offset);

			if(socket_x_offset > this_box_width) {
				this_box_width = socket_x_offset;
			}
			this_box_height += socket_y_offset;
			draw_HB_connect_space(cr,cur_mod,cur_xcor,this_box_width-2*D_SPACE);
		}

		draw_box(cr,cur_xcor,cur_ycor,this_box_width,this_box_height,false,cur_mod->mod);
		if(!args_format.is_bw_only())color_module_dependent_box(cr,cur_xcor,cur_ycor,this_box_width,this_box_height,&box_draw_bounds,color_palet,colorer,cur_mod->mod);
		
		if(args_format.is_drawing_ports()) draw_channels_wrapper(cr,cur_mod->mod,cur_xcor,cur_ycor,this_box_width,temp_ycur_offset,VisualPortVector);


		if(this_box_deepest_level > param_deepest_level) {
			param_deepest_level = this_box_deepest_level;
		}

		cur_xcor -= D_SPACE;
		cur_ycor -= D_SPACE;
		cur_level--;
		args_format.unset_print_tag(cur_mod->mod->get_name());
		}
		height_tracker = this_box_height;
		return this_box_width;
	}
}

void toggle_color(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_bw();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_quarter(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(0.25);
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_half(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(0.5);
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_full(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(1);
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_double(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(2);
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_increment(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->increment_scale_factor_by5percent();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_decrement(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->decrement_scale_factor_by5percent();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_increment(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->increment_depth();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_decrement(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->decrement_depth();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_toggle(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_depth();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void port_toggle(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter * args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_ports();
	args_formatter->pending_update = true;

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void ExposedEvent(GtkWidget *widget,GdkEvent *event, gpointer* data) {

		risc::Design* d = (risc::Design*) data[0]; 
		int deepest_level = 1;
		int top_width = 0;
		int top_height = 0;
		static int count = 0;
		std::vector<VisualPort> top_ports;
	
		ArgumentFormatter* pnt_passer = (ArgumentFormatter*) data[2];
		ArgumentFormatter& args_format = (*pnt_passer);

		if(args_format.pending_update) {
			args_format.pending_update = false;
		
			global_drawn_sockets.clear();
			cairo_cmd_list.clear();	
	
			std::vector<VisualDrawnSocket*> submod_drawn_HB_sockets;
			std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash> sub_mod_HB_map;
			top_width = draw_recursor(top_,deepest_level,top_height,top_ports,d->mapping_info_,(cairo_t*)1,args_format,submod_drawn_HB_sockets,sub_mod_HB_map);
		
			if(top_width != global_unscaled_top_width || top_height != global_unscaled_top_height) {
				global_unscaled_top_width = top_width;
				global_unscaled_top_height = top_height;
				gtk_widget_set_size_request(widget,top_width * args_format.get_scale_factor() + 2 * D_SPACE,top_height * args_format.get_scale_factor() + 2 * D_SPACE);
			}

			if(surface != NULL) {
				cairo_surface_destroy(surface);
			}
			surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
				global_unscaled_top_width +2*D_SPACE, global_unscaled_top_height + 2*D_SPACE);
		
			cairo_t* layer = cairo_create(surface);
			cairo_scale(layer,args_format.get_scale_factor(),args_format.get_scale_factor());
			cairo_set_source_rgb(layer,0.1,0.1,0.1);
			cairo_select_font_face(layer,font_name,CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
			cairo_set_font_size(layer,10);
			for(std::list< std::pair< std::string, VisualCommandArg > >::iterator cmd_iter =  cairo_cmd_list.begin();
				cmd_iter != cairo_cmd_list.end();
				cmd_iter++) {
				(*cairo_commands.at(cmd_iter->first))(layer,cmd_iter->second);
			}	
			cairo_destroy(layer);
		}

		
		cairo_t *cr = gdk_cairo_create(widget->window);	
		cairo_set_source_surface(cr, surface, 0, 0);
		cairo_paint(cr);
		cairo_destroy(cr);
	
}

void ExposedEventTestMode(GtkWidget *widget,GdkEvent *event, gpointer* data) {

		risc::Design* d = static_cast<risc::Design*>(data[0]); 
		int deepest_level = 1;
		int top_width = 0;
		int top_height = 0;
		static int count = 0;
		std::vector<VisualPort> top_ports;
	
		ArgumentFormatter* pnt_passer = static_cast<ArgumentFormatter*>( data[2]);
		ArgumentFormatter& args_format = (*pnt_passer);

		if(args_format.pending_update) {
			args_format.pending_update = false;
		
			global_drawn_sockets.clear();
			cairo_cmd_list.clear();	
	
			std::vector<VisualDrawnSocket*> submod_drawn_HB_sockets;
			std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash> sub_mod_HB_map;
			top_width = draw_recursor(top_,deepest_level,top_height,top_ports,d->mapping_info_,(cairo_t*)1,args_format,submod_drawn_HB_sockets,sub_mod_HB_map);
		
			if(top_width != global_unscaled_top_width || top_height != global_unscaled_top_height) {
				global_unscaled_top_width = top_width;
				global_unscaled_top_height = top_height;
				// gtk_widget_set_size_request(widget,top_width * args_format.get_scale_factor() + 2 * D_SPACE,top_height * args_format.get_scale_factor() + 2 * D_SPACE);
			}

			if(surface != NULL) {
				cairo_surface_destroy(surface);
			}
			surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
				global_unscaled_top_width +2*D_SPACE, global_unscaled_top_height + 2*D_SPACE);
		
			cairo_t* layer = cairo_create(surface);
			cairo_scale(layer,args_format.get_scale_factor(),args_format.get_scale_factor());
			cairo_set_source_rgb(layer,0.1,0.1,0.1);
			cairo_select_font_face(layer,font_name,CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
			cairo_set_font_size(layer,10);
			for(std::list< std::pair< std::string, VisualCommandArg > >::iterator cmd_iter =  cairo_cmd_list.begin();
				cmd_iter != cairo_cmd_list.end();
				cmd_iter++) {
				(*cairo_commands.at(cmd_iter->first))(layer,cmd_iter->second);
			}	
			cairo_destroy(layer);
		}

		std::cout << "Generating " << global_input_file_name << ".png" << std::endl;
		cairo_t *cr = cairo_create(surface);
		cairo_set_source_surface(cr, surface, 0, 0);
		cairo_paint(cr);
		cairo_surface_write_to_png(surface, strcat(global_input_file_name, ".png"));
	
		
}

void draw_page(GtkPrintOperation* operation, GtkPrintContext* context, gint pgNr,gpointer* data) {
	risc::Design* d = (risc::Design*) data[0]; 	

	int deepest_level = 1;
	int top_width = 0;
	int top_height;
	g_print("draw_page function called\n");
	std::vector<VisualPort> top_ports;

	ArgumentFormatter* pnt_passer = (ArgumentFormatter*) data[2];
	ArgumentFormatter& args_format = (*pnt_passer);
	
	cairo_t* cr = gtk_print_context_get_cairo_context(context);
	
	cairo_scale(cr,args_format.get_scale_factor(),args_format.get_scale_factor());

	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_select_font_face(cr,font_name,CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size(cr,10);

	/* OLD VERSION
	for(risc::InstanceVectorIter iter = d->top_modules_.begin();
		iter != d->top_modules_.end();
		iter++ ) {
		top_width += draw_recursor((*iter),deepest_level,top_height,top_ports,top_sockets,d->mapping_info_,cr,args_format);
	}*/
	std::vector<VisualDrawnSocket*> submod_drawn_HB_sockets;
	std::unordered_map<std::pair<int,SgSymbol*>, int ,pairhash> sub_mod_HB_map;
	top_width = draw_recursor(top_,deepest_level,top_height,top_ports,d->mapping_info_,cr,args_format,submod_drawn_HB_sockets,sub_mod_HB_map);
	cairo_scale(cr,args_format.get_scale_factor(),args_format.get_scale_factor());
	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_select_font_face(cr,font_name,CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size(cr,10);
	for(std::list< std::pair< std::string, VisualCommandArg > >::iterator cmd_iter =  cairo_cmd_list.begin();
		cmd_iter != cairo_cmd_list.end();
		cmd_iter++) {
		(*cairo_commands.at(cmd_iter->first))(cr,cmd_iter->second);
	}

}

void do_print(GtkWidget* widget, gpointer* data) {
	std::cout << "In do_print!\n";

	GtkPrintOperation* print;
	GtkPrintSettings* settings;
	GtkPaperSize* paper_size;

	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	print = gtk_print_operation_new();
	gtk_print_operation_set_unit(print,GTK_UNIT_POINTS);
	gtk_print_operation_set_n_pages(print,1);
	gtk_print_operation_set_use_full_page(print,TRUE);
	gtk_print_operation_set_export_filename(print,"visualization.pdf");

	g_signal_connect(print,"draw_page",G_CALLBACK(draw_page),data);	

	paper_size = gtk_paper_size_new_custom("Visualization","display_Visualiztion",
						global_unscaled_top_width * args_formatter->get_scale_factor() + 2 * D_SPACE,global_unscaled_top_height * args_formatter->get_scale_factor() + 2 * D_SPACE,GTK_UNIT_POINTS);
	settings = gtk_print_settings_new();
	gtk_print_settings_set_paper_size(settings,paper_size);
	gtk_print_operation_set_print_settings(print,settings);



	gtk_print_operation_run(print,GTK_PRINT_OPERATION_ACTION_EXPORT,NULL,NULL);

}

int main(int argc, char* argv[]) {
	global_input_file_name = argv[1];
	if(argc > 2 && std::strcmp(argv[2], "-test:visual") == 0) 
	{
		global_test_mode = true;
	}




	cairo_commands["set_line_width"] = visual_cairo_set_line_width;
	cairo_commands["set_source_rgb"] = visual_cairo_set_source_rgb;
	cairo_commands["move_to"] = visual_cairo_move_to;
	cairo_commands["line_to"] = visual_cairo_line_to;
	cairo_commands["rectangle"] = visual_cairo_rectangle;
	cairo_commands["fill"] = visual_cairo_fill;
	cairo_commands["set_source_rgba"] = visual_cairo_set_source_rgba;
	cairo_commands["set_line_cap"] = visual_cairo_set_line_cap;
	cairo_commands["stroke"] = visual_cairo_stroke;
	cairo_commands["set_font_size"] = visual_cairo_set_font_size;
	cairo_commands["show_text"] = visual_cairo_show_text;
	cairo_commands["curve_to"] = visual_cairo_curve_to;
	surface = NULL;
	
	GtkWidget *window;
	GtkWidget *scrolled_window;
	GtkWidget * DrawArea;

	GtkWidget *menubar;
	GtkWidget *viewmenu;
	GtkWidget *view;
	GtkWidget *tog_stat;
	GtkWidget *zoomMenu;
	GtkWidget *zoom;
	GtkWidget *halfzoom;
	GtkWidget *quarterzoom;
	GtkWidget *fullzoom;
	GtkWidget *doublezoom;
	GtkWidget *sep;
	GtkWidget *incrementzoom;
	GtkWidget *decrementzoom;
	GtkWidget *depthmenu;
	GtkWidget *depth;
	GtkWidget *depthadd;
	GtkWidget *depthdecrease;
	GtkWidget *depthtoggle;	
	GtkWidget *toggleport;

	GtkWidget *file;
	GtkWidget *filemenu;
	GtkWidget *quit;
	GtkWidget *exportjpg;
	GtkWidget *print_btn;	

	void* pnt_buffer[4];

	ArgumentFormatter args_formatter(argc,argv);

	// add -DRISC (RD, 07/06/18)
	risc::tools::AddCommandlineOption(argc, argv, "-DRISC");

	std::string LWIncludePath;
	LWIncludePath = "-I";
	LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
	LWIncludePath += "/include";
	
	// add -I$SYSTEMC_LW_HOME/include
	risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);

	std::cout << std::endl
            	<< "Processing File..."
            	<< std::endl;

	risc::Design d(argc, argv);
	//std::cout << "created design" << std::endl;
	pnt_buffer[0] = &d;

	std::cout << std::endl
		<< "Finished Processing File."
		<< std::endl;
	//visual initla walk
	top_ = initial_walk(d.top_modules_.front());
	//std::cout << "Finished Initial Walk!\n";
	//check_HB_sockets(top_);
	//std::cout << "\n";
	//check_ports(top_);	

	pnt_buffer[2] = &args_formatter;
	//convert unordered mapping to ordered for test, ZC 2019 DEC 11
	risc::MappingInfo::TreeMap& itm 
		= risc::get_design()->mapping_info_.inst_tree_mapping;
	
	for(risc::MappingInfo::TreeMap::iterator 
		it = itm.begin();
		it != itm.end();
		++it)
	{
		ordered_tree_mapping[it->first] = it->second;
	}

	if(global_test_mode)
	{
		ExposedEventTestMode(NULL,NULL,pnt_buffer);
		return 0;
	}
	gtk_init(&argc, &argv);
	window = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(window),"Module Visualization");
	gtk_window_set_default_size(GTK_WINDOW(window),300,300);
	gtk_window_set_resizable(GTK_WINDOW(window),TRUE);
	gtk_window_set_type_hint(GTK_WINDOW(window),GDK_WINDOW_TYPE_HINT_NORMAL);	
	pnt_buffer[1] = window;

	menubar = gtk_menu_bar_new();
	viewmenu = gtk_menu_new();
	filemenu = gtk_menu_new();
	zoomMenu = gtk_menu_new();
	depthmenu = gtk_menu_new();

	file = gtk_menu_item_new_with_label("File");
	print_btn = gtk_menu_item_new_with_label("Print");
	quit = gtk_menu_item_new_with_label("Quit");

	view = gtk_menu_item_new_with_label("View");
	quarterzoom = gtk_menu_item_new_with_label("25%");
	halfzoom = gtk_menu_item_new_with_label("50%");
	fullzoom = gtk_menu_item_new_with_label("100%");
	doublezoom = gtk_menu_item_new_with_label("200%");
	zoom = gtk_menu_item_new_with_label("Zoom");
	sep = gtk_separator_menu_item_new();
	incrementzoom = gtk_menu_item_new_with_label("Increment by 5%");
	decrementzoom = gtk_menu_item_new_with_label("Decrement by 5%");
	depth = gtk_menu_item_new_with_label("Depth");
	depthadd = gtk_menu_item_new_with_label("Increment Depth Level");
	depthdecrease = gtk_menu_item_new_with_label("Decrement Depth Level");
	depthtoggle = gtk_check_menu_item_new_with_label("Depth Variation Enabled");

	tog_stat = gtk_check_menu_item_new_with_label("No Color");
	if(args_formatter.is_bw_only()) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tog_stat), TRUE);
	}
	else {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tog_stat), FALSE);
	}

	toggleport = gtk_check_menu_item_new_with_label("Draw Channels and Ports");
	if(args_formatter.is_drawing_ports()) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(toggleport),TRUE);
	}
	else {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(toggleport),FALSE);
	}

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(file),filemenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(filemenu),print_btn);
	gtk_menu_shell_append(GTK_MENU_SHELL(filemenu),quit);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), file);

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(view),viewmenu);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(zoom),zoomMenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),quarterzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),halfzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),fullzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),doublezoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),sep);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),incrementzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),decrementzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),zoom);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(depth),depthmenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthadd);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthdecrease);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthtoggle);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),depth);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu), tog_stat);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),toggleport);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), view);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), menubar, FALSE, FALSE, 0);


	scrolled_window = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),GTK_POLICY_AUTOMATIC,GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox),scrolled_window,TRUE,TRUE,0);

	DrawArea = gtk_drawing_area_new();
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),DrawArea);
	pnt_buffer[3] = DrawArea;

	gtk_widget_show_all(window);
	
	g_signal_connect(G_OBJECT(DrawArea),"expose_event",G_CALLBACK(ExposedEvent),pnt_buffer);

	g_signal_connect(G_OBJECT(window),"destroy",G_CALLBACK(gtk_main_quit),NULL);

	g_signal_connect(G_OBJECT(quit),"activate",G_CALLBACK(gtk_main_quit),NULL);

	g_signal_connect(G_OBJECT(tog_stat), "activate", G_CALLBACK(toggle_color),pnt_buffer);

	g_signal_connect(G_OBJECT(print_btn), "activate", G_CALLBACK(do_print),pnt_buffer);

	g_signal_connect(G_OBJECT(quarterzoom),"activate",G_CALLBACK(zoom_quarter),pnt_buffer);

	g_signal_connect(G_OBJECT(halfzoom),"activate",G_CALLBACK(zoom_half),pnt_buffer);

	g_signal_connect(G_OBJECT(fullzoom),"activate",G_CALLBACK(zoom_full),pnt_buffer);

	g_signal_connect(G_OBJECT(doublezoom),"activate",G_CALLBACK(zoom_double),pnt_buffer);

	g_signal_connect(G_OBJECT(incrementzoom),"activate",G_CALLBACK(zoom_increment),pnt_buffer);
	
	g_signal_connect(G_OBJECT(decrementzoom),"activate",G_CALLBACK(zoom_decrement),pnt_buffer);

	g_signal_connect(G_OBJECT(depthadd),"activate",G_CALLBACK(depth_increment),pnt_buffer);

	g_signal_connect(G_OBJECT(depthdecrease),"activate",G_CALLBACK(depth_decrement),pnt_buffer);

	g_signal_connect(G_OBJECT(depthtoggle),"activate",G_CALLBACK(depth_toggle),pnt_buffer);

	g_signal_connect(G_OBJECT(toggleport),"activate",G_CALLBACK(port_toggle),pnt_buffer);

	gtk_main();	
}
