#ifndef PORT_MAPPING_H_INCLUDED_
#define PORT_MAPPING_H_INCLUDED_

#include <list>

#include "rose.h"

#include "instance_tree.h"

namespace risc {

class Port;
class Module;

/**
 * \fn       get_mapped_channel_instances(Port *port)
 * \brief    This function determines to which channels this port can be bound.
 *           An instance of the given port can occur multiple times.
 */
std::list<InstanceTree> get_mapped_channel_instances(Port *port);

/**
 * \fn       get_mapped_channel_instance(InstanceTree port)
 * \brief    This function determines to which channel this port is bound
 */
InstanceTree get_mapped_channel_instance(InstanceTree port);

/**
 * \fn       get_mapped_channel_instance_for_psg(InstanceTree port)
 * \brief    This function determines to which channel this port is bound
 *           given mapping information in a PSG file
 */
InstanceTree get_mapped_channel_instance_for_psg(InstanceTree port);

/**
 * \fn       port_mapping_style1(
                SgFunctionCallExp *func_call,
                SgVariableSymbol *port,
                SgVariableDefinition *parent_variable)
 * \brief    Determines if function call bindes in style of
 *           sub_module.port.bind(port1);
 * \return   Symbol pointer to port1
 */
SgVariableSymbol* port_mapping_style1(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    SgVariableDefinition *parent_variable);

/**
 * \fn       port_mapping_style2(
                SgFunctionCallExp *func_call,
                SgVariableSymbol *port,
                SgVariableDefinition *parent_variable)
 * \brief    Determines if function call bindes in style of
 *           sub_module.port(port1);
 * \return   Symbol pointer to port1
 */
SgVariableSymbol* port_mapping_style2(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    SgVariableDefinition *parent_variable);

/**
 * \fn       port_mapping_style3(
                SgFunctionCallExp *func_call,
                SgVariableSymbol *port,
                Module *parent_module,
                SgVariableDefinition *parent_variable)
 * \brief    Determines if function call bindes in style of
 *           sub_module(port1, port2, ... ,portn);
 */
SgVariableSymbol* port_mapping_style3(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable);

/**
 * \fn       socket_mapping_style2(
                SgFunctionCallExp *func_call,
                SgVariableSymbol *socket,
                SgVariableDefinition *parent_variable)
 * \brief    Determines if function call bindes in style of
 *           sub_module1.portx(sub_module2.porty);
 */
SgVariableSymbol* socket_mapping_style2(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *socket,
    SgVariableDefinition *parent_variable);

/**
 * \fn       get_mapped_symbol(
                SgVariableSymbol *port,
                Module *parent_module,
                SgVariableDefinition *parent_variable)
 * \brief    This function is only for internal use
 *           Here we determine to which object (port or channel) the given
 *           port is mapped in the parent module
 * \param    port            Port declartion in the sub module
 * \param    parent_module   Parent module top
 * \param    parent_variable Declaration of the sub module in Top module
 * \return   bounded channel/port something
 *           Example: submodule.port.bind(something);
 */
SgVariableSymbol* get_mapped_symbol(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable);

/**
 * \fn       get_mapped_symbol_style3(
                SgVariableSymbol *port,
                Module *parent_module,
                SgVariableDefinition *parent_variable)
 * \brief    This function is only for internal use
 *           Here we determine to which object (port or channel) the given
 *           port is mapped in the parent module
 * \param    port            Port declartion in the sub module
 * \param    parent_module   Parent module top
 * \param    parent_variable Declaration of the sub module in Top module
 * \return   bounded channel/port something
 *           Example: submodule.port.bind(something);
 */
SgVariableSymbol* get_mapped_symbol_style3(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable);

/**
 * \fn       get_mapped_module_instance_symbol(
                SgVariableSymbol *socket,
                Module *parent_module,
                SgVariableDefinition *parent_variable)
 * \brief    This function is only for internal use
 *           Here we determine to which module instance the given
 *           socket is mapped in the parent module
 * \param    socket          Socket declartion in the sub module
 * \param    parent_module   Parent module top
 * \param    parent_variable Declaration of the sub module in parent module
 * \return   bounded module instance var symbol something
 *           Example: submodule.socket_this.bind(something.socket_another);
 */
SgVariableSymbol* get_mapped_module_instance_symbol(
    SgVariableSymbol *socket,
    Module *parent_module,
    SgVariableDefinition *parent_variable);

/**
 * \fn       module_instance_mapping_style1(
                SgFunctionCallExp *func_call,
                SgVariableSymbol *socket,
                SgVariableDefinition *parent_variable)
 * \brief    Determines if function call bindes in style of
 *           parent_variable.socket(some_module_instance.socket_another);
 * \return   Symbol pointer to some_module_instance
 */
SgVariableSymbol* module_instance_mapping_style1(
    SgFunctionCallExp *func_call,
    SgVariableSymbol *socket,
    SgVariableDefinition *parent_variable);

    
/**
 * \fn       get_mapped_socket_instance_tree(InstanceTree instance_tree_1)
 * \brief    This function determines to which socket this socket is bound
 */
risc::InstanceTree get_mapped_socket_instance_tree(InstanceTree instance_tree_1);
} // end of namespace risc

#endif /* PORT_MAPPING_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
