#include "segment.h"

#include "function_annotation_attributes.h"
#include "cached_function_ast_attributes.h"
#include "reference_attributes.h"
#include "generic_payload_attributes.h"
#include "../internal_representation/design.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/function_helper.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"
#include "../tools/systemc_definitions.h"

// RD: EXPERIMENT using Rose to obtain read-write access
//#include "variables.h" // (for getReadVarRefsInSubtree)

int risc::sg::Segment::id_counter = 0;
int risc::sg::Segment::node_counter = 0;
//int risc::sg::Segment::id_counter_offset = -1;

risc::sg::Segment::Segment(
    SgNode *boundary_node,
    bool inlined_segment,
    int psg_id,
    bool is_fake_segment,
    bool is_simulation_process):
  id_(Segment::id_counter),
  boundary_node_(boundary_node),
  originating_thread_(NULL),
  calls_conflict_function_(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched),
  inlined_segment_(inlined_segment),
  deref_pointer_read_(false),
  deref_pointer_write_(false),
  contains_sc_stop_(false),
  is_channel_segment_(false),
  is_b_transport_seg_(false),
  b_transport_func_(NULL),
  b_transport_(NULL),
  wait_offset_(-1),
  
  node_id_(node_counter),
  psg_id_(psg_id),
  function_name_entrance_(""),
  function_name_exit_(""),
  non_defining_function_call_name_(""),
  pcp_port_name_(""),
  pcp_port_(NULL),
  is_fake_segment_(is_fake_segment),
  is_sc_fifo_port_call_(false),
  is_simulation_process_(is_simulation_process),
  segment_node_id_ (0),
  module_name_ (""),
  is_entrance_to_channel_method_ (false),
  port_call_func_exp_(NULL),
  unique_port_call_name_(""),
  channel_seg_ids_array_name_(""),
  is_redacted_(false),
  min_waiting_time_(0,0, true)

{
 
  /*
    under PSG mode, segment id for fake segment node
    and non-defining function call node
    should not increase, because this will affect the 
    instrumented segment id
  */
  Segment::node_counter++;
  if(PSG_MODE){    
    if(!is_fake_segment_){    
      Segment::id_counter++;
    }   
  }
  else{
    Segment::id_counter++;    
  }
  
  debugging_ = true;
  waits_for_time_ = waits_for_time();
}

risc::sg::Segment::Segment(const Segment &other):
  id_(other.id_),
  node_id_(other.node_id_),
  boundary_node_(other.boundary_node_),
  originating_thread_(other.originating_thread_),
  expressions_(other.expressions_),
  calls_conflict_function_(other.calls_conflict_function_),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched),
  deref_pointer_read_(other.deref_pointer_read_),
  deref_pointer_write_(other.deref_pointer_write_),
  contains_sc_stop_(other.contains_sc_stop_),

  inlined_segment_(other.inlined_segment_), //fixed

  waits_for_time_(other.waits_for_time_),
  is_channel_segment_(other.is_channel_segment_),
  is_b_transport_seg_(other.is_b_transport_seg_),
  b_transport_func_(other.b_transport_func_),
  b_transport_(other.b_transport_),
  wait_offset_(other.wait_offset_),
  pcp_(other.pcp_),
  
  read_variables_(other.read_variables_),
  write_variables_(other.write_variables_),
  wait_events_(other.wait_events_),
  notify_events_(other.notify_events_),
  events_notification_time_(other.events_notification_time_),
  psg_id_(other.psg_id_),

  function_name_entrance_(other.function_name_entrance_),
  function_name_exit_(other.function_name_exit_),
  non_defining_function_call_name_(other.non_defining_function_call_name_),
  pcp_port_name_(other.pcp_port_name_),
  pcp_port_(other.pcp_port_),
  is_fake_segment_(other.is_fake_segment_),
  is_sc_fifo_port_call_(other.is_sc_fifo_port_call_),
  is_simulation_process_(other.is_simulation_process_),
  segment_node_id_(other.segment_node_id_),
  module_name_(other.module_name_),
  is_entrance_to_channel_method_(other.is_entrance_to_channel_method_),
  port_call_func_exp_(other.port_call_func_exp_),
  unique_port_call_name_(other.unique_port_call_name_),
  channel_seg_ids_array_name_(other.channel_seg_ids_array_name_),
  is_redacted_(other.is_redacted_),
  min_waiting_time_(other.min_waiting_time_)
{
  debugging_ = true;
}

bool
risc::sg::Segment::operator==(const Segment &other)
{
  return id_ == other.id_;
}

void
risc::sg::Segment::add_expression(NodeWithPath nwp)
{
  if(nwp.node_) {
    // add all expressions in a channel to the conflict analysis
    expressions_.push_back(nwp);
  }
}

void
risc::sg::Segment::merge(const Segment &other)
{
  if(*this == other)
    return;

  calls_conflict_function_ |= this->calls_conflict_function_;
  deref_pointer_read_ |= this->deref_pointer_read_;
  deref_pointer_write_ |= this->deref_pointer_write_;
  contains_sc_stop_ |= this->contains_sc_stop_;

  for(Expressions::const_iterator
      iter  = other.expressions_.begin();
      iter != other.expressions_.end();
      iter++) {
    add_expression(*iter);
  }
  
  std::map<SgVariableSymbol*, bool> var_and_flag;

  for(std::set<SymbolWithPath>::const_iterator
      iter  = read_variables_.begin();
      iter != read_variables_.end();
      iter++) {
    var_and_flag[(*iter).symbol_] = true;
  }

  for(std::set<SymbolWithPath>::const_iterator
      iter  = other.read_variables_.begin();
      iter != other.read_variables_.end();
      iter++) {
    if(var_and_flag[(*iter).symbol_] == false){
        read_variables_.insert((*iter));
        var_and_flag[(*iter).symbol_] = true;
      }
  }
  
  var_and_flag.clear();
  for(std::set<SymbolWithPath>::const_iterator
      iter  = write_variables_.begin();
      iter != write_variables_.end();
      iter++) {
    var_and_flag[(*iter).symbol_] = true;
  }

  for(std::set<SymbolWithPath>::const_iterator
      iter  = other.write_variables_.begin();
      iter != other.write_variables_.end();
      iter++) {
    if(var_and_flag[(*iter).symbol_] == false){
        write_variables_.insert((*iter));
        var_and_flag[(*iter).symbol_] = true;
    } 
  }
  
  for(std::set<SymbolWithPath>::const_iterator
      iter  = other.wait_events_.begin();
      iter != other.wait_events_.end();
      iter++) {
    wait_events_.insert(*iter);
  }
  
  for(std::set<SymbolWithPath>::const_iterator
      iter  = other.notify_events_.begin();
      iter != other.notify_events_.end();
      iter++) {
    notify_events_.insert(*iter);
  }
  
  for(std::map<SgVariableSymbol*, Time>::const_iterator
      iter  = other.events_notification_time_.begin();
      iter != other.events_notification_time_.end();
      iter++) {
    events_notification_time_[iter->first] = iter->second;
  }
  
}

void
risc::sg::Segment::analyze_segment()  // perform RW-analysis for a segment
{                                     // (revised/fixed 04/26/19, RD)
//#define DEBUG // (04/24/19, RD)
// (for reference, see test/regression/data_conflict3.cc)

#ifdef DEBUG
  printf("\nDEBUG: Segment::analyze_segment(ID %d)\n", id_);
#endif
  for(Segment::Expressions::iterator
      iter  = expressions_.begin();
      iter != expressions_.end();
      iter++) {

    SgExprStatement *stmt_expr;
    SgExpression *expr;
    SgVariableDeclaration *var_decl;

    if((stmt_expr = isSgExprStatement(iter->node_)))
    { // example: '((this)->Port)->write(p);'
#ifdef DEBUG
      printf("DEBUG: STMT %s\n", stmt_expr->unparseToString().c_str());
#endif
      analyze_expression(stmt_expr->get_expression(), NonRead, NonWrite, iter->pcp_);
    }
    else if((expr = isSgExpression(iter->node_)))
    {
      SgIfStmt *IfStmnt;
      SgSwitchStatement *SwitchStmnt;
      SgWhileStmt *WhileStmnt;
      SgDoWhileStmt *DoWhileStmnt;
      SgForStatement *ForStmnt;
      SgNode *Parent = expr->get_parent();
      SgNode *GrandParent = Parent->get_parent();

      if(isSgReturnStmt(Parent))
      { // example: return 'temp'
#ifdef DEBUG
        printf("DEBUG: RTRN %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else if(  (IfStmnt = isSgIfStmt(GrandParent))
              &&(IfStmnt->get_conditional()==Parent))
      { // example: if ('cond') ...
#ifdef DEBUG
        printf("DEBUG: IFEL %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else if(  (SwitchStmnt = isSgSwitchStatement(GrandParent))
              &&(SwitchStmnt->get_item_selector()==Parent))
      { // example: switch ('val') ...
#ifdef DEBUG
        printf("DEBUG: SWTC %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else if(  (WhileStmnt = isSgWhileStmt(GrandParent))
              &&(WhileStmnt->get_condition()==Parent))
      { // example: while ('cond') ...
#ifdef DEBUG
        printf("DEBUG: WHIL %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else if(  (DoWhileStmnt = isSgDoWhileStmt(GrandParent))
              &&(DoWhileStmnt->get_condition()==Parent))
      { // example: do { } while('cond')
#ifdef DEBUG
        printf("DEBUG: DOWH %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else if(  (ForStmnt = isSgForStatement(GrandParent))
              &&(ForStmnt->get_test()==Parent))
      { // example: for (...; 'test'; ...) ...
#ifdef DEBUG
        printf("DEBUG: FORC %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, Read, NonWrite, iter->pcp_);
      }
      else
      { // example: 'x++' or '(this)->v2 = false' or '(this)->ack.notify(SC_ZERO_TIME)'
#ifdef DEBUG
        printf("DEBUG: EXPR %s\n", expr->unparseToString().c_str());
#endif
        analyze_expression(expr, NonRead, NonWrite, iter->pcp_);
      }
    }
    else if((var_decl = isSgVariableDeclaration(iter->node_)))
    { // example: 'int Local;' or 'int temp =(this)->tmpd;'
#ifdef DEBUG
      printf("DEBUG: DECL %s\n", var_decl->unparseToString().c_str());
#endif
      SgInitializedName *init_name = var_decl->get_decl_item(0);
      if(init_name->get_initptr())
      {
        write_variables_.insert(
            SymbolWithPath(
              isSgVariableSymbol(
                init_name->search_for_symbol_from_symbol_table()),
              iter->pcp_));
        /*
        generic_payload_analysis_helper(
          init_name->search_for_symbol_from_symbol_table(),
          true)
        */
        analyze_expression(init_name->get_initptr(), NonRead, NonWrite, iter->pcp_);
      }
    }
    else
    { // example: 'break;' or 'continue;' or ';'
#ifdef DEBUG
      printf("DEBUG: OTHR %s\n", iter->node_->unparseToString().c_str());
#endif
    }
  }

#if 0 // RD: BEGIN EXPERIMENT using Rose to obtain read-write access
  for(Segment::Expressions::iterator
      iter  = expressions_.begin();
      iter != expressions_.end();
      iter++) {
    varIDSet ReadVars = getReadVarRefsInSubtree(iter->node_);
// The line above fails due to a missing "rwAccess" annotation!
// edg4x-rose/src/midend/programAnalysis/genericDataflow/rwAccessLabeler/rwAccessLabeler.C:332:
// rwAccessLabeler::accessType rwAccessLabeler::getAccessType(SgNode*):
// Assertion `n->getAttribute("rwAccess")' failed.
// This happens for the simple expression "(this) -> a1 = 42".
// I give up here, try to fix Tim's approach instead!
// RD, 06/29/18.
    for(std::set<varID>::iterator VarIter = ReadVars.begin();
        VarIter != ReadVars.end();
        VarIter++) {
      printf("DEBUG(getReadVar): %s\n", VarIter->str().c_str());
    }
  }
#endif // RD: END EXPERIMENT using Rose to obtain read-write access
}

// Helper function for identifying class-variable assignments (RD, 06/29/18)
inline int isOverloadedAssignmentOperator(
    SgFunctionCallExp *func_call_exp)
{
  if(func_call_exp->get_args()->get_expressions().size() != 1) {
    return(0);
  }
  if (!func_call_exp->get_uses_operator_syntax()) {
    return(0);
  }
  SgDotExp *DotExp = isSgDotExp(func_call_exp->get_function());
  if (!DotExp) {
    return(0);
  }
  SgMemberFunctionRefExp *Operator = isSgMemberFunctionRefExp(DotExp->get_rhs_operand());
  if (!Operator) {
    return(0);
  }
  std::string Op = Operator->unparseToString();
  if (Op == "=") {
    return(1); // regular assignment operator
  }
  if (  (Op == "+=")
      ||(Op == "-=")
      ||(Op == "*=")
      ||(Op == "/=")
      ||(Op == "%=")
      ||(Op == "<<=")
      ||(Op == ">>=")
      ||(Op == "&=")
      ||(Op == "|=")
      ||(Op == "^=")) {
    return(2); // augmented assignment operator (aka. compound assignment operator)
  }
  return(0);
}

void risc::sg::Segment::analyze_expression(
    SgNode *expression, ERead read, EWrite write, PortCallPath pcp)
{
  if(expression == NULL) {
    return;
  }


  #define ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(Operator)              \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), Read, NonWrite, pcp); \
    analyze_expression(op->get_lhs_operand(), NonRead, Write, pcp); \
    analyze_expression(op->get_rhs_operand(), Read, NonWrite, pcp); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgAndAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgIorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgXorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgLshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgRshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgPlusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMinusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMultAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgDivAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgModAssignOp);
  #undef ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP

  #define ANALYZE_VARS_IN_BINARY_OP(Operator)                       \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), Read, NonWrite, pcp); \
    analyze_expression(op->get_rhs_operand(), Read, NonWrite, pcp); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_BINARY_OP(SgLessThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLessOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgEqualityOp);
  ANALYZE_VARS_IN_BINARY_OP(SgNotEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLshiftOp);
  ANALYZE_VARS_IN_BINARY_OP(SgRshiftOp);

  ANALYZE_VARS_IN_BINARY_OP(SgBitAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitXorOp);

  ANALYZE_VARS_IN_BINARY_OP(SgAddOp);
  ANALYZE_VARS_IN_BINARY_OP(SgSubtractOp);
  ANALYZE_VARS_IN_BINARY_OP(SgDivideOp);
  ANALYZE_VARS_IN_BINARY_OP(SgMultiplyOp);
  ANALYZE_VARS_IN_BINARY_OP(SgModOp);

  ANALYZE_VARS_IN_BINARY_OP(SgCommaOpExp);
  #undef ANALYZE_VARS_IN_BINARY_OP

  #define ANALYZE_VARS_IN_UNARY_OP(Operator)                    \
  if(is##Operator(expression)) {                                \
    Operator *op = is##Operator(expression);                    \
    analyze_expression(op->get_operand(), Read, NonWrite, pcp); \
    return;                                                     \
  }
  ANALYZE_VARS_IN_UNARY_OP(SgUnaryAddOp);
  ANALYZE_VARS_IN_UNARY_OP(SgMinusOp);
  ANALYZE_VARS_IN_UNARY_OP(SgAssignInitializer);
  ANALYZE_VARS_IN_UNARY_OP(SgNotOp);
  ANALYZE_VARS_IN_UNARY_OP(SgBitComplementOp);
  ANALYZE_VARS_IN_UNARY_OP(SgCastExp);
  #undef ANALYZE_VARS_IN_UNARY_OP

  #define ANALYZE_VARS_IN_UNARY_INC_DEC_OP(Operator)         \
  if(is##Operator(expression)) {                             \
    Operator *op = is##Operator(expression);                 \
    analyze_expression(op->get_operand(), Read, Write, pcp); \
    return;                                                  \
  }
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgPlusPlusOp);
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgMinusMinusOp);
  #undef ANALYZE_VARS_IN_UNARY_INC_DEC_OP

  #define ANALYZE_VARS_IN_ACCESS_OP(Operator)                    \
  if(is##Operator(expression)) {                                 \
    Operator *op = is##Operator(expression);                     \
    analyze_expression(op->get_rhs_operand(), Read, write, pcp); \
    analyze_expression(op->get_lhs_operand(), Read, write, pcp); \
    return;                                                      \
  }
  ANALYZE_VARS_IN_ACCESS_OP(SgDotStarOp);
  ANALYZE_VARS_IN_ACCESS_OP(SgArrowStarOp);
  #undef ANALYZE_VARS_IN_ACCESS_OP



  if(isSgArrowExp(expression)) {
    SgArrowExp *op = isSgArrowExp(expression);

    analyze_expression(op->get_lhs_operand(), Read, write, pcp);

    if(write) {
      analyze_expression(op->get_rhs_operand(), read, Write, pcp);
    } else {
      analyze_expression(op->get_rhs_operand(), read, NonWrite, pcp);
    }
    return;
  }

  if(isSgDotExp(expression)) {
    SgDotExp *op = isSgDotExp(expression);

    if(read && write) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite, pcp);
      analyze_expression(op->get_rhs_operand(), Read, Write, pcp);
    } else if(write) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite, pcp);
      analyze_expression(op->get_rhs_operand(), NonRead, Write, pcp);
    } else if(read) {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite, pcp);
      analyze_expression(op->get_rhs_operand(), Read, NonWrite, pcp);
    } else {

      analyze_expression(op->get_lhs_operand(), NonRead, NonWrite, pcp);
      analyze_expression(op->get_rhs_operand(), NonRead, NonWrite, pcp);
    }

    return;
  }

  if(isSgSizeOfOp(expression)) {
    SgSizeOfOp *size_of_op = isSgSizeOfOp(expression);
    analyze_expression(size_of_op->get_operand_expr(), Read, NonWrite, pcp);
    return;
  }

  if(isSgVarRefExp(expression)) {

    SgVarRefExp *var_ref = isSgVarRefExp(expression);

    risc::tools::analyze_conflict_free(var_ref->get_symbol());

    if(read) {
      read_variables_.insert(SymbolWithPath(var_ref->get_symbol(), pcp));

      generic_payload_analysis_helper(var_ref->get_symbol(), false, pcp);
    } else {

      SgModifierType *modifier = isSgModifierType(var_ref->get_type());

      if(modifier) {

        if(modifier->get_typeModifier().get_constVolatileModifier().
            isVolatile()) {

          if(!write) {
            read_variables_.insert(
                SymbolWithPath(var_ref->get_symbol(), pcp));
            generic_payload_analysis_helper(var_ref->get_symbol(), false, pcp);
          }
        }
      }
    }

    if(write) {

      write_variables_.insert(SymbolWithPath(var_ref->get_symbol(), pcp));
      
      generic_payload_analysis_helper(var_ref->get_symbol(), true, pcp);
    }
    return;
  }

  if(isSgFunctionCallExp(expression)) {

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expression);

// BEGIN: HANDLE OVERLOADED OPERATORS FOR ASSIGNMENTS OF CLASS-TYPE VARIABLES (RD, 06/29/18)
    int OpType = isOverloadedAssignmentOperator(func_call_exp);
    if (OpType > 0) {
      if (OpType == 1) { // regular assignment operator

      // emulate the case for SgAssignOp:

      // analyze_expression(assign_op->get_lhs_operand(), read, Write, pcp);
      analyze_expression(isSgDotExp(func_call_exp->get_function())->get_lhs_operand(),
                          read, Write, pcp);

      // analyze_expression(assign_op->get_rhs_operand(), Read, NonWrite, pcp);
      analyze_expression(func_call_exp->get_args()->get_expressions().at(0),
                          Read, NonWrite, pcp);
      } else { // augmented assignment operator (aka. compound assignment operator)

      // emulate the cases for SgCompoundAssignOp:

      // analyze_expression(op->get_lhs_operand(), Read, NonWrite, pcp);
      analyze_expression(isSgDotExp(func_call_exp->get_function())->get_lhs_operand(),
                          Read, NonWrite, pcp);

      // analyze_expression(op->get_lhs_operand(), NonRead, Write, pcp);
      analyze_expression(isSgDotExp(func_call_exp->get_function())->get_lhs_operand(),
                          NonRead, Write, pcp);

      // analyze_expression(op->get_rhs_operand(), Read, NonWrite, pcp);
      analyze_expression(func_call_exp->get_args()->get_expressions().at(0),
                          Read, NonWrite, pcp);
      }
      return;
    }
// END: HANDLE OVERLOADED OPERATORS FOR ASSIGNMENTS OF CLASS-TYPE VARIABLES (RD, 06/29/18)

    SgExpressionPtrList expr_list
      = func_call_exp->get_args()->get_expressions();

    // Here we have to check if any parameter type is a pointer.
    // If so, the segment is with all other segments in conflict
    SgFunctionDeclaration *func_decl
      = tools::get_first_non_defining_function_declaration(func_call_exp);

    if(func_decl) 
    {
      //zc Sept 20 2019
      //first check if the function call is sc_stop,
      //if yes, simply set the contains_sc_stop_ flag;
      if(func_decl->get_symbol_from_symbol_table() == 
          risc::tools::SystemCDefinitions::get_sc_stop()->get_symbol_from_symbol_table())
      {
        contains_sc_stop_ = true;
        return;
      }

      //ZC, 2018/12/20 hack for memcpy
      if(SageInterface::get_name(func_decl) == "memcpy")
      {
        //std::cout << "is memcpy" << std::endl;
        //std::cout << "func_call_exp = " << func_call_exp->unparseToString() << std::endl;
        std::vector<SgNode*> var_refs = 
              NodeQuery::querySubTree(func_call_exp, V_SgVarRefExp);
        for(std::vector<SgNode*>::iterator
            var_ref_iter  = var_refs.begin();
            var_ref_iter != var_refs.end();
            var_ref_iter ++) 
        {
          SgVarRefExp *var_ref = isSgVarRefExp(*var_ref_iter);
          if(var_ref){
            //std::cout << "var_ref is " << var_ref->get_symbol()->get_name() << std::endl;
            //SgVariableSymbol* trans = 
            //syms_lhs.push_back(var_ref->get_symbol());
          }
        }
        
        SgVariableSymbol* dest = isSgVariableSymbol( isSgVarRefExp(var_refs[0])->get_symbol());
        write_variables_.insert(
                SymbolWithPath(dest, pcp));
        generic_payload_analysis_helper(dest, true, pcp);

        SgVariableSymbol* src = isSgVariableSymbol( isSgVarRefExp(var_refs[1])->get_symbol());
        read_variables_.insert(
                SymbolWithPath(src, pcp));
        generic_payload_analysis_helper(src, false, pcp);

        if(var_refs.size()>=3)
        {
          SgVariableSymbol* len = isSgVariableSymbol( isSgVarRefExp(var_refs[2])->get_symbol());
          read_variables_.insert(SymbolWithPath(len, pcp));
          generic_payload_analysis_helper(len, false, pcp);
        }
        return;
      } 
      else if(SageInterface::get_name(func_decl) == "next_trigger")
      {
        if(func_call_exp->get_args()->get_expressions().size() == 1) {
              
          std::vector<SgNode*> var_refs = 
              NodeQuery::querySubTree(func_call_exp, V_SgVarRefExp);
          for(std::vector<SgNode*>::iterator
              var_ref_iter  = var_refs.begin();
              var_ref_iter != var_refs.end();
              var_ref_iter ++) 
          {
            SgVarRefExp *var_ref = isSgVarRefExp(*var_ref_iter);
            if(var_ref){
              //SgVariableSymbol* trans = 
              //syms_lhs.push_back(var_ref->get_symbol());
              if(risc::tools::is_sc_event(var_ref)) {
                wait_events_.insert(
                      SymbolWithPath(var_ref->get_symbol(),pcp_));

              }
            }
          }

          // if(isSgArrowExp(func_call_exp->get_args()->
          //         get_expressions().at(0))){
          //   SgVarRefExp* waiting_event
          //     = isSgVarRefExp(isSgArrowExp(func_call_exp->get_args()->
          //           get_expressions().at(0))->get_rhs_operand());
          //   if(risc::tools::is_sc_event(waiting_event)) {
          //     std::cout << "event is " << waiting_event->unparseToString() << std::endl;
          //     wait_events_.insert(
          //           SymbolWithPath(waiting_event->get_symbol(),pcp_));

          //   }
          // }
        }
      }
      FunctionAnnotationAttributes *function_annotation_attribute =
          dynamic_cast<FunctionAnnotationAttributes*>(
              func_decl->getAttribute("function_annotation"));

      if(function_annotation_attribute) {

        if(!function_annotation_attribute->is_conflict_free_) {
          calls_conflict_function_ = true;

          Conflict conflict(func_call_exp);
          conflicts_.insert(conflict);
        }
      }

      int argument_counter = 0;
      SgExprListExp *arguments = func_call_exp->get_args();
      for(SgExpressionPtrList::iterator
          expr_iter  = expr_list.begin();
          expr_iter != expr_list.end();
          expr_iter++, argument_counter++) {

        if(argument_counter
            < func_decl->get_parameterList()->get_args().size()) {

          SgReferenceType *ref_type
            = isSgReferenceType(func_decl->get_parameterList()
                ->get_args().at(argument_counter)->get_type());

          if(ref_type) {

            analyze_expression(*expr_iter, NonRead, NonWrite, pcp);

          } else {

            analyze_expression(*expr_iter, Read, NonWrite, pcp);

          }
        } else { // this function signature has the ... parameter

            analyze_expression(*expr_iter, Read, NonWrite, pcp);
        }
      }
    }
    return;
  }

  if(isSgPntrArrRefExp(expression)) {
    SgPntrArrRefExp *pointer_array_ref = isSgPntrArrRefExp(expression);

    if(read && write) {
      analyze_expression(
          pointer_array_ref->get_lhs_operand(), Read, Write, pcp);
      analyze_expression(
          pointer_array_ref->get_rhs_operand(), Read, NonWrite, pcp);

    } else if(write) {
      analyze_expression(
          pointer_array_ref->get_lhs_operand(), NonRead, Write, pcp);
      analyze_expression(
          pointer_array_ref->get_rhs_operand(), Read, NonWrite, pcp);

    } else if(read) {
      analyze_expression(
          pointer_array_ref->get_lhs_operand(), Read, NonWrite, pcp);
      analyze_expression(
          pointer_array_ref->get_rhs_operand(), Read, NonWrite, pcp);

    } else {
      analyze_expression(
          pointer_array_ref->get_lhs_operand(), NonRead, NonWrite, pcp);
      analyze_expression(
          pointer_array_ref->get_rhs_operand(), Read, NonWrite, pcp);
    }

    return;
  }

  if(isSgAssignOp(expression)) {
    SgAssignOp *assign_op = isSgAssignOp(expression);

    analyze_expression(assign_op->get_lhs_operand(), read, Write, pcp);

    analyze_expression(assign_op->get_rhs_operand(), Read, NonWrite, pcp);

    return;
  }

  if(isSgConditionalExp(expression)) {
    SgConditionalExp *conditional_exp = isSgConditionalExp(expression);

    analyze_expression(
        conditional_exp->get_conditional_exp(), Read, NonWrite, pcp);

    analyze_expression(conditional_exp->get_true_exp(), Read, NonWrite, pcp);

    analyze_expression(conditional_exp->get_false_exp(), Read, NonWrite, pcp);

    return;
  }

  if(isSgPointerDerefExp(expression)) {
    //std::cout << "pointer!!!!!!!!!!!!" << std::endl;
    SgPointerDerefExp *op = isSgPointerDerefExp(expression);

    analyze_expression(op->get_operand(), Read, NonWrite, pcp);

    if(read) {
      deref_pointer_read_ = true;
    }

    if(write) {
      deref_pointer_write_ = true;
    }

    if(debugging_) {
      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  if(isSgAddressOfOp(expression)) {
    SgAddressOfOp *op = isSgAddressOfOp(expression);

    analyze_expression(op->get_operand(), NonRead, NonWrite, pcp);

    if(debugging_) {
      Conflict conflict(expression);
      conflicts_.insert(conflict);
    }

    return;
  }

  assert(true && "Operator not supported");
  return;
}

static bool RD_EvalConstExpr(       // better than SageInterface, but still not good
              SgExpression *expr,   // fix for bug20190621.cc (08/21/19, RD)
              long double *value)
{
  SgExpression *ValueExpr;

  // skip compiler-generated cast-expression to expected 'double' type
  if (isSgCastExp(expr)) {
    ValueExpr = isSgCastExp(expr)->get_operand();
  } else {
    ValueExpr = expr;
  }

  // get floating-point values (which this bug fix is about)
  if (isSgFloatVal(ValueExpr)) {
    *value = isSgFloatVal(ValueExpr)->get_value();
    return true;
  } else if (isSgDoubleVal(ValueExpr)) {
    *value = isSgDoubleVal(ValueExpr)->get_value();
    return true;
  } else if (isSgLongDoubleVal(ValueExpr)) {
    *value = isSgLongDoubleVal(ValueExpr)->get_value();
    return true;
  }

  // use SageInterface for integer constants
  SageInterface::const_int_expr_t IntValue =
    SageInterface::evaluateConstIntegerExpression(ValueExpr);
  if(IntValue.hasValue_) {
    *value = IntValue.value_;
    return true;
  }

  // more complex cases are not supported at this time
  *value = -1.0;
  return false;
}

risc::sg::Time
risc::sg::Segment::get_min_waiting_time()
{
  // TimeAdvanceTable::get_waiting_time_for_segment() and
  // Segment::get_min_waiting_time()
  // were very similar (with a lot of code duplication);
  // both functions are now combined (thanks to ZC!);
  // result of the analysis below is stored with each segment
  // so that it can be looked up (without re-calculation)
  // (07/10/18 and 08/21/19, RD)

  if(!boundary_node_) {
      return Time(0,0);
     // initial segment
  }
  if(min_waiting_time_.infinite_ == false) //initialized
  {
    return min_waiting_time_; // look-up previously calculated value (ZC)
  }
  min_waiting_time_.infinite_ = false;
  const SgStatement *stmt = isSgStatement(boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    const SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        SgExpressionPtrList &function_call_exprs
          = func_call->get_args()->get_expressions();

        // In this case we can wait for
        //   - sc_event_and_list, sc_event_or_list
        //   - sc_event
        //   - sc_zero_time
        //   - (time, time unit)

        if(function_call_exprs.size() == 0) {
          // wait(/*nothing*/) // wait on static-sensitivity list (or forever)
          if(risc::get_design()->no_immediate_notify)
            min_waiting_time_ = risc::sg::Time(0,1);
          else
            min_waiting_time_ = risc::sg::Time(0,0);

          return min_waiting_time_; // wait for at least a delta-cycle
          // RD: if static sensitivity is empty, we could return "infinity" here?

        } else if(function_call_exprs.size() == 1) {

          std::string arg0_string(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString());
          // SC_ZERO_TIME
          if(arg0_string == "SC_ZERO_TIME") {
            min_waiting_time_ = risc::sg::Time(0,1);
            return min_waiting_time_; // wait for exactly one delta-cycle

          } else if(arg0_string == "((this) -> simcontext())") {
            
            if(risc::get_design()->no_immediate_notify)
              min_waiting_time_ = risc::sg::Time(0,1);
            else
              min_waiting_time_ = risc::sg::Time(0,0);
            return min_waiting_time_; //DM 07/23/2019 for sc_method

          }
          // TODO: wait(const_value_of_type_sc_time)
          // TODO: wait(const_value_of_type_int)
          else {
            //DM 6/4/2018
            SgArrowExp* arrow_exp = isSgArrowExp(function_call_exprs.at(0));
            if(arrow_exp) {
              // SC_EVENT
              SgVarRefExp* waiting_event = isSgVarRefExp(arrow_exp->get_rhs_operand());
              if(risc::tools::is_sc_event(waiting_event) ||
                 risc::tools::is_sc_event_and_list(waiting_event) ||
                 risc::tools::is_sc_event_or_list(waiting_event)) 
              {
                if(risc::get_design()->no_immediate_notify)
                  min_waiting_time_ = risc::sg::Time(0,1);
                else
                  min_waiting_time_ = risc::sg::Time(0,0);
                return min_waiting_time_; // wait for at least a delta-cycle
              }
            }
            else {
                std::string or_list_event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 16 ? 
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-16,16) : "";

                std::string event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 8 ?
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-8,8) : "";

                std::string and_list_event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 17 ?
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-17,17) : "";

                if(event_string == "sc_event"
                    || or_list_event_string == "sc_event_or_list"
                    || and_list_event_string == "sc_event_and_list") 
                {
                  if(risc::get_design()->no_immediate_notify)
                    min_waiting_time_ = risc::sg::Time(0,1);
                  else
                    min_waiting_time_ = risc::sg::Time(0,0);
                  return min_waiting_time_;
                }
            }
            //end code DM
          }
        } else if(function_call_exprs.size() == 2 || function_call_exprs.size() == 3 ) {
          // something like wait(1,SC_NS)

          long double Value;
          if (RD_EvalConstExpr(function_call_exprs.at(0), &Value)) {
            if (Value == 0.0) {
              min_waiting_time_ = risc::sg::Time(0,1);
              return min_waiting_time_;
            }
            min_waiting_time_ = Time(Value *
                Time::convert_si_magnitude(function_call_exprs.at(1)
                  ->unparseToString()),0);
            return min_waiting_time_;

          } else {
            // wait(an expression cannot be evaluated), ZC
            // just treat it as ZERO_TIME
            // std::cout << "Warning: Wait should have a constant expression."
            //           << std::endl;
            // std::cout << "       See line "
            //           << func_call->get_file_info()->get_line()
            //           << " in file "
            //           << func_call->get_file_info()->get_filename()
            //           << std::endl;
            if(risc::get_design()->no_immediate_notify)
              min_waiting_time_ = risc::sg::Time(0,1);
            else
              min_waiting_time_ = risc::sg::Time(0,0);
            // exit(1);
          }

        } else {
          std::cout << "Error: Wait time cannot be determined."
                    << std::endl;
          std::cout << "        See line "
                    << func_call->get_file_info()->get_line()
                    << " in file "
                    << func_call->get_file_info()->get_filename()
                    << std::endl;
          exit(1);
        }
      }
    }
  }

  if(risc::get_design()->no_immediate_notify)
    min_waiting_time_ = risc::sg::Time(0,1);
  else
    min_waiting_time_ = risc::sg::Time(0,0);  // identification fails, return worst case
  return min_waiting_time_;
}

bool risc::sg::Segment::waits_for_time()
{
  if(!boundary_node_) {
    return false;
  }

  const SgStatement *stmt = isSgStatement(boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        SgExpressionPtrList &function_call_exprs
          = func_call->get_args()->get_expressions();

        // In this case we can wait for
        //   - sc_event_and_list, sc_event_or_list
        //   - sc_event
        //   - sc_zero_time
        //   - (time, time unit)
        //
        if(function_call_exprs.size() == 0) {

          return false; //DM 6/1/2018

        } else if(function_call_exprs.size() == 1) {

          std::string arg0_string(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString());
          // SC_ZERO_TIME
          if(arg0_string == "SC_ZERO_TIME") {
            return true;

          } else if(arg0_string == "((this) -> simcontext())") {
              return false; //DM 07/23/2019
          }
          // TODO: wait(const_value_of_type_sc_time)
          // TODO: wait(const_value_of_type_int)
          else {
            //DM 6/1/2018
            SgArrowExp* arrow_exp = isSgArrowExp(function_call_exprs.at(0));
            if(arrow_exp) {
              // SC_EVENT
              SgVarRefExp* waiting_event = isSgVarRefExp(arrow_exp->get_rhs_operand());
              if(risc::tools::is_sc_event(waiting_event) ||
                 risc::tools::is_sc_event_and_list(waiting_event) ||
                 risc::tools::is_sc_event_or_list(waiting_event)) {
                return false;
              }
            }
            else {
                std::string or_list_event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 16 ? 
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-16,16) : "";

                std::string event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 8 ?
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-8,8) : "";

                std::string and_list_event_string = SageInterface::get_name(function_call_exprs.at(0)->get_type()).size() >= 17 ?
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).substr(
                  SageInterface::get_name(function_call_exprs.at(0)->get_type()).size()-17,17) : "";

                if(event_string == "sc_event"
                    || or_list_event_string == "sc_event_or_list"
                    || and_list_event_string == "sc_event_and_list") {
                      return false;
                }

                //end DM code
            }

          }
        } else if(function_call_exprs.size() == 2 || function_call_exprs.size() == 3) {
          // something like wait(1,SC_NS)

          long double dont_care;
          // Hack for wait(everyting) for now, ZC
          return true;
          if (RD_EvalConstExpr(function_call_exprs.at(0), &dont_care)) {
            return true;
          } else { 

            std::cout << "Error: Wait must have a constant expression."
                      << std::endl;
            std::cout << "       See line "
                      << func_call->get_file_info()->get_line()
                      << " in file "
                      << func_call->get_file_info()->get_filename()
                      << std::endl;
            exit(1);
          }

        } else {

          std::cout << "Error: Wait time cannot be determined."
                    << std::endl;
          std::cout << "        See line "
                    << func_call->get_file_info()->get_line()
                    << " in file "
                    << func_call->get_file_info()->get_filename()
                    << std::endl;
          exit(1);
        }
      }
    }
  }

  return false;
}

void risc::sg::Segment::generic_payload_analysis_helper(SgVariableSymbol* var_sym, bool is_write, PortCallPath pcp_of_var)
{
  //added for generic payload reference analysis, ZC, 2018,12,15
  
  //1st, get the target side gp object of the var
  Tlm2ContainerOfMappedDataAttribute *mapped_container_attribute =
    dynamic_cast<Tlm2ContainerOfMappedDataAttribute*>(
      var_sym->getAttribute("tlm2_container_of_mapped_data_attribute"));
  if(mapped_container_attribute) 
  {
    /*std::cout << "\ngeneric payload reference analysis in dataconflicttable.cpp" <<std::endl;
    if(is_write) std::cout << "adding to write_variables_" << std::endl;
    else std::cout << "adding to read_variables_" << std::endl;
    std::cout << "var is " << var_sym->get_name() << std::endl;*/
    SgVariableSymbol* container = mapped_container_attribute->container_;
    if(container)
    {
      //std::cout << "container is " << container->get_name() << std::endl;
      //2nd, decide if this is a dmi object or a gp object
      MappedContainersOfContainerAttribute *mapped_containers_attribute =
        dynamic_cast<MappedContainersOfContainerAttribute*>(
          container->getAttribute("mapped_containers_of_container_attribute"));

      if(mapped_containers_attribute) 
      {
        for(std::unordered_map<SgVariableSymbol*, std::list<SgVariableDefinition*> >::iterator
            mapping_iter  = mapped_containers_attribute->symbols_sockets_.begin();
            mapping_iter != mapped_containers_attribute->symbols_sockets_.end();
            ++ mapping_iter
          )
        {

          PortCallPath pcp;
          SgVariableSymbol* mapped_container = (*mapping_iter).first;
          

          std::list<SgVariableDefinition*> path = (*mapping_iter).second;

          OriginalDataAttribute *original_data_attribute =
            dynamic_cast<OriginalDataAttribute*>(
              mapped_container->getAttribute("original_data_attribute"));
          if(original_data_attribute == NULL) continue;
          SgVariableSymbol* original_data = original_data_attribute->data_symbol_;

          //3rd, decide if container is a dmi object
          DmiSocketAttribute *dmi_socket_attribute =
            dynamic_cast<DmiSocketAttribute*>(
                container->getAttribute("dmi_socket_attribute"));

          //if it is a dmi object, then we simply add all the mapped original data
          //to read_/write_variables
          if(dmi_socket_attribute) {
            /*std::cout << "mapped_container is "     
                    << mapped_container->get_name() << std::endl;
            std::cout << "original_data is "     
                    << original_data->get_name() << std::endl;*/
            //old version
            /*SgVariableDefinition* socket = dmi_socket_attribute->socket_;
            pcp.pcp_type_ = 1;
            pcp.port_.push_front(socket);*/
            pcp = PortCallPath(path);
            /*std::cout << "original_data has pcp socket for dmi analysis: " 
              << socket->get_declaration()->get_decl_item(NULL)->get_qualified_name() 
              << std::endl;*/

            if(is_write) write_variables_.insert(SymbolWithPath(original_data, pcp));
            else read_variables_.insert(SymbolWithPath(original_data, pcp));

          }
          //otherwise, it is a gp object.
          //we need to find the container that has a matching pcp with var_sym
          else
          {
            pcp = PortCallPath(path);
            if(pcp == pcp_of_var) //matching
            {

              /*std::cout << "mapped_container is "     
                      << mapped_container->get_name() << std::endl;
              std::cout << "original_data is "     
                      << original_data->get_name() << std::endl;
              std::cout << "**found matching pcp for generic payload" << std::endl;
              pcp.print();*/
              if(is_write) write_variables_.insert(
                SymbolWithPath(original_data, PortCallPath())
              );
              else read_variables_.insert(
                SymbolWithPath(original_data, PortCallPath())
              );
                break;
            }
          }
        }
      }
    }
  } //end of generic payload reference analysis, ZC, 2018,12,15
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
