// particle_channel5w.cc: regression test for event synchronization
// (originated from NoC particle simulation example with lost events)
// -> here 3 modules communicate in a cycle, clock- and counter-clock-wise
// RD, 08/02/17

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS (42*10)

// user-defined data type

typedef struct Particle_t
{
  double 	x, y;

  inline friend std::ostream& operator<<(
	std::ostream& os, const Particle_t& co)
  { os << co.x << ' ' << co.y << '\n';
    return os;
  }
} particle_t;

// user-defined typed double-handshake channel (with transceiver interface)

class i_particle_transceiver: virtual public sc_interface
{
    public:	// (mimic a subset of sc_fifo interface, without references)
	virtual particle_t read(void) = 0;
	virtual void write(const particle_t) = 0;
};

class c_particle_double_handshake_channel:
	public sc_channel, public i_particle_transceiver
{
    public:
	c_particle_double_handshake_channel(sc_module_name name):
		sc_channel(name)
	{
	    v = false;
	    w = false;
	}

	~c_particle_double_handshake_channel()
	{ }

	particle_t read(void)
	{
	    if (!v)
	    {
		w = true;
		wait(req);
		w = false;
	    }
	    particle_t temp = tmpd;
	    v = false;
	    ack.notify(SC_ZERO_TIME);
	    wait(ack);
	    return temp;
	}

	void write(const particle_t X)
	{
	    tmpd = X;
	    v = true;
	    if (w)
	    {
		req.notify(SC_ZERO_TIME);
	    }
	    wait(ack);
	}

    private:
	sc_event req;
	sc_event ack;
	bool v, w;
	particle_t tmpd;
};

// Round Robin East-West

SC_MODULE(RoundRobinEW)
{
    sc_port<i_particle_transceiver> PortE;
    sc_port<i_particle_transceiver> PortW;
    unsigned Count;

    void main(void);

    SC_CTOR(RoundRobinEW)
    {
	Count = 0;
	SC_HAS_PROCESS(RoundRobinEW);
	SC_THREAD(main);
    }
};

void RoundRobinEW::main(void)
{
    particle_t p;
    int i;

    for(i=0; i<ITERATIONS; i++)
    {
        p.x = 1.2;
        p.y = 3.4;
	wait(1, SC_NS);
	PortE->write(p);
	wait(1, SC_NS);
	p = PortE->read();
	assert(p.x == 1.2);
	assert(p.y == 3.4);
	wait(1, SC_NS);
	p = PortW->read();
	wait(1, SC_NS);
	PortW->write(p);
	Count++;
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Round Robin West-East

SC_MODULE(RoundRobinWE)
{
    sc_port<i_particle_transceiver> PortW;
    sc_port<i_particle_transceiver> PortE;
    unsigned Count;

    void main(void);

    SC_CTOR(RoundRobinWE)
    {
	Count = 0;
	SC_HAS_PROCESS(RoundRobinWE);
	SC_THREAD(main);
    }
};

void RoundRobinWE::main(void)
{
    particle_t p;
    int i;

    for(i=0; i<ITERATIONS; i++)
    {
	wait(1, SC_NS);
	p = PortW->read();
	wait(1, SC_NS);
	PortW->write(p);
        p.x = 1.2;
        p.y = 3.4;
	wait(1, SC_NS);
	PortE->write(p);
	wait(1, SC_NS);
	p = PortE->read();
	assert(p.x == 1.2);
	assert(p.y == 3.4);
	Count++;
    }
    assert(i == ITERATIONS);
    assert(Count == ITERATIONS);
}

// Top

SC_MODULE(TOP)
{
    c_particle_double_handshake_channel c1, c2, c3;
    RoundRobinEW rr1;
    RoundRobinWE rr2;
    RoundRobinEW rr3;

    SC_CTOR(TOP)
    :c1("c1"), c2("c2"), c3("c3"), rr1("rr1"), rr2("rr2"), rr3("rr3")
    {
    }

    void before_end_of_elaboration()
    {
	rr1.PortE.bind(c1);
	rr1.PortW.bind(c3);
	rr2.PortE.bind(c2);
	rr2.PortW.bind(c1);
	rr3.PortE.bind(c3);
	rr3.PortW.bind(c2);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("Round-robin cycles starts\n");
    sc_start();
//  printf("Round-robin counts: %d %d %d.\n",
//	top.rr1.Count, top.rr2.Count, top.rr3.Count);
    assert(top.rr1.Count == ITERATIONS);
    assert(top.rr2.Count == ITERATIONS);
    assert(top.rr3.Count == ITERATIONS);
    printf("Round-robin cycles done.\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
