// timed_pipe.cc: timed pipeline with 0 inner stages
// (inspired from delta_pipe.cc)
// 05/02/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define BUFSIZE    2
#define STAGES     0
#define ITEMS      10
#define WORKLOAD   10000000*STAGES
#define WORKTIME   5.0
//#define VERBOSE

#pragma RISC M conflict-free
int M[STAGES+1][BUFSIZE] = { }; // memory for double-buffering

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> OutNumSent;

    void main(void)
    {
	for(int i=0; i<ITEMS; i++)
	{
	    // make an item
	    wait(WORKTIME, SC_PS);
	    Count++;
	    wait(WORKTIME, SC_PS);

	    // if buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Source waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in buffer and notify receiver
	    M[Buffer][Index] = Count;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), Count);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next send
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*ITEMS));
    }

    Source(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};


SC_MODULE(Sink)
{
    int Buffer, Index, d, Count;
    sc_in<unsigned>  InNumSent;
    sc_out<unsigned> InNumRcvd;

    void main(void)
    {
	for(int i=0; i<ITEMS; i++)
	{
	    // if buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Sink waiting for item S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load items from buffer and notify sender
	    d = M[Buffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next receive

	    // check the item
	    wait(WORKTIME, SC_PS);
	    Count++;
	    assert(d == Count);
	    wait(WORKTIME, SC_PS);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    }

    Sink(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_signal<unsigned> NumSent1, NumRcvd1;

    Source src;
    Sink   snk;

    SC_CTOR(TOP)
    : NumSent1("NumSent1", 0),
	NumRcvd1("NumRcvd1", 0),
	src("src", 0),
	snk("snk", 0)
    {
	src.OutNumSent.bind(NumSent1);
	src.OutNumRcvd.bind(NumRcvd1);
	snk.InNumSent.bind(NumSent1);
	snk.InNumRcvd.bind(NumRcvd1);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("timed_pipe starts\n");
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp().value() == (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    printf("timed_pipe done.\n");
    printf("[potential speedup over SEQ: OOO 1x, NPD 1x, SYN 1x]\n");
    return 0;
}

// EOF
