// timed_pipe3x3_rnd2.cc: timed pipeline with 3x3 inner stages and random workload
// (same as timed_pipe3x3_rnd.cc, but with decentralized WaitAnyTime)
// 05/20/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define BUFSIZE    2
#define LANES      3
#define STAGES     3
#define ITEMS      9
#define WORKLOAD   10000000*STAGES*LANES
#define WORKTIME   5
//#define VERBOSE

#pragma RISC M conflict-free
int M[LANES][STAGES+1][BUFSIZE] = { }; // memory for double-buffering

int Rnd(unsigned long &Next, int Max)
{
    Next = Next * 1103515245 + 12345;
    return((unsigned)(Next/65536) % (Max+1));
}

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_in<unsigned>  OutNumRcvdA;
    sc_in<unsigned>  OutNumRcvdB;
    sc_in<unsigned>  OutNumRcvdC;
    sc_out<unsigned> OutNumSentA;
    sc_out<unsigned> OutNumSentB;
    sc_out<unsigned> OutNumSentC;

    void WaitAnyTime(unsigned long &RndNext, int MaxPicos)
    {
	int picos = Rnd(RndNext, MaxPicos+1)-1; // range [-1...MaxPicos]

	if(picos >= 0)	// don't wait if picos is negative;
	{   // wait for a delta-cycle or some time
	    // printf("DEBUG WaitAnyTime waiting for %d pico seconds.\n", picos);
	    wait((double)picos, SC_PS);
	}
    else
	{   // skip (no wait)
	    // printf("DEBUG WaitAnyTime NOT waiting.\n");
	}
    }

    void main(void)
    {
	unsigned long RndNext = 0;

	for(int i=0; i<ITEMS; i++)
	{
	    // make an item
	    WaitAnyTime(RndNext, WORKTIME);
	    Count++;
	    WaitAnyTime(RndNext, WORKTIME);

	    // if buffer is full, wait for free space
	    while (  (OutNumSentA - OutNumRcvdA >= BUFSIZE)
		   ||(OutNumSentB - OutNumRcvdB >= BUFSIZE)
		   ||(OutNumSentC - OutNumRcvdC >= BUFSIZE))
	    {
//		printf("%2u:%02u: DEBUG Source waiting for output buffer space A=%u, B=%u, C=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSentA - OutNumRcvdA, OutNumSentB - OutNumRcvdB, OutNumSentC - OutNumRcvdC);
		wait(  OutNumRcvdA.value_changed_event()
		     | OutNumRcvdB.value_changed_event()
		     | OutNumRcvdC.value_changed_event());
	    }

	    // store item in buffer and notify receiver
	    for(int l=0; l<LANES; l++)
	    {
		M[l][Buffer][Index] = (l+1)*10+Count;
	    }
	    OutNumSentA = OutNumSentA + 1;
	    OutNumSentB = OutNumSentB + 1;
	    OutNumSentC = OutNumSentC + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), Count);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next send
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*ITEMS));
    }

    Source(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};

SC_MODULE(Stage)
{
    int Lane, InBuffer, OutBuffer, Index, d;
    sc_in<unsigned>  InNumSent;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> InNumRcvd;
    sc_out<unsigned> OutNumSent;
    unsigned long w;

    void WaitAnyTime(unsigned long &RndNext, int MaxPicos)
    {
	int picos = Rnd(RndNext, MaxPicos+1)-1; // range [-1...MaxPicos]

	if(picos >= 0)	// don't wait if picos is negative;
	{   // wait for a delta-cycle or some time
	    // printf("DEBUG WaitAnyTime waiting for %d pico seconds.\n", picos);
	    wait((double)picos, SC_PS);
	}
    else
	{   // skip (no wait)
	    // printf("DEBUG WaitAnyTime NOT waiting.\n");
	}
    }

    void main(void)
    {
	unsigned long RndNext = (Lane+1)*10+InBuffer;

	for(;;)
	{
	    // if input buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for input buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load item from buffer and notify sender
	    d = M[Lane][InBuffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction

	    // process the item
#ifdef VERBOSE
	    printf("%2u:%02u: %s processes data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    WaitAnyTime(RndNext, WORKTIME);
	    w = Work(Rnd(RndNext, WORKLOAD));
	    WaitAnyTime(RndNext, WORKTIME);

	    // if output buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in output buffer and notify receiver
	    M[Lane][OutBuffer][Index] = d;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction
	}
    }

    Stage(sc_module_name _nm, int _l, int _ib, int _ob)
    :	sc_module(_nm), Lane(_l), InBuffer(_ib), OutBuffer(_ob)
    {
	Index = 0;
	SC_HAS_PROCESS(Stage);
	SC_THREAD(main);
    }
};

SC_MODULE(Sink)
{
    int Buffer, Index, d[LANES], Count;
    sc_in<unsigned>  InNumSentA;
    sc_in<unsigned>  InNumSentB;
    sc_in<unsigned>  InNumSentC;
    sc_out<unsigned> InNumRcvdA;
    sc_out<unsigned> InNumRcvdB;
    sc_out<unsigned> InNumRcvdC;

    void WaitAnyTime(unsigned long &RndNext, int MaxPicos)
    {
	int picos = Rnd(RndNext, MaxPicos+1)-1; // range [-1...MaxPicos]

	if(picos >= 0)	// don't wait if picos is negative;
	{   // wait for a delta-cycle or some time
	    // printf("DEBUG WaitAnyTime waiting for %d pico seconds.\n", picos);
	    wait((double)picos, SC_PS);
	}
    else
	{   // skip (no wait)
	    // printf("DEBUG WaitAnyTime NOT waiting.\n");
	}
    }

    void main(void)
    {
	unsigned long RndNext = 1;

	for(int i=0; i<ITEMS; i++)
	{
	    // if buffer is empty, wait for items
	    while (  (InNumSentA - InNumRcvdA == 0)
		   ||(InNumSentB - InNumRcvdB == 0)
		   ||(InNumSentC - InNumRcvdC == 0))
	    {
//		printf("%2u:%02u: DEBUG Sink waiting for items A=%u, B=%u, C=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSentA - InNumRcvdA, InNumSentB - InNumRcvdB, InNumSentC - InNumRcvdC);
		wait(  InNumSentA.value_changed_event()
		     | InNumSentB.value_changed_event()
		     | InNumSentC.value_changed_event());
	    }

	    // load items from buffer and notify sender
	    for(int l=0; l<LANES; l++)
	    {
		d[l] = M[l][Buffer][Index];
	    }
	    InNumRcvdA = InNumRcvdA + 1;
	    InNumRcvdB = InNumRcvdB + 1;
	    InNumRcvdC = InNumRcvdC + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d[0]%10);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next receive

	    // check the item
	    WaitAnyTime(RndNext, WORKTIME);
	    Count++;
	    for(int l=0; l<LANES; l++)
	    {
		assert(d[l] == (l+1)*10+Count);
	    }
	    WaitAnyTime(RndNext, WORKTIME);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    }

    Sink(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_signal<unsigned> NumSentA1, NumRcvdA1;
    sc_signal<unsigned> NumSentB1, NumRcvdB1;
    sc_signal<unsigned> NumSentC1, NumRcvdC1;
    sc_signal<unsigned> NumSentA2, NumRcvdA2;
    sc_signal<unsigned> NumSentB2, NumRcvdB2;
    sc_signal<unsigned> NumSentC2, NumRcvdC2;
    sc_signal<unsigned> NumSentA3, NumRcvdA3;
    sc_signal<unsigned> NumSentB3, NumRcvdB3;
    sc_signal<unsigned> NumSentC3, NumRcvdC3;
    sc_signal<unsigned> NumSentA4, NumRcvdA4;
    sc_signal<unsigned> NumSentB4, NumRcvdB4;
    sc_signal<unsigned> NumSentC4, NumRcvdC4;

    Source src;
    Stage  sA1, sB1, sC1;
    Stage  sA2, sB2, sC2;
    Stage  sA3, sB3, sC3;
    Sink   snk;

    SC_CTOR(TOP)
    : NumSentA1("NumSentA1", 0),
	NumRcvdA1("NumRcvdA1", 0),
	NumSentB1("NumSentB1", 0),
	NumRcvdB1("NumRcvdB1", 0),
	NumSentC1("NumSentC1", 0),
	NumRcvdC1("NumRcvdC1", 0),
	NumSentA2("NumSentA2", 0),
	NumRcvdA2("NumRcvdA2", 0),
	NumSentB2("NumSentB2", 0),
	NumRcvdB2("NumRcvdB2", 0),
	NumSentC2("NumSentC2", 0),
	NumRcvdC2("NumRcvdC2", 0),
	NumSentA3("NumSentA3", 0),
	NumRcvdA3("NumRcvdA3", 0),
	NumSentB3("NumSentB3", 0),
	NumRcvdB3("NumRcvdB3", 0),
	NumSentC3("NumSentC3", 0),
	NumRcvdC3("NumRcvdC3", 0),
	NumSentA4("NumSentA4", 0),
	NumRcvdA4("NumRcvdA4", 0),
	NumSentB4("NumSentB4", 0),
	NumRcvdB4("NumRcvdB4", 0),
	NumSentC4("NumSentC4", 0),
	NumRcvdC4("NumRcvdC4", 0),
	src("src", 0),
	sA1("sA1", 0, 0, 1),
	sB1("sB1", 1, 0, 1),
	sC1("sC1", 2, 0, 1),
	sA2("sA2", 0, 1, 2),
	sB2("sB2", 1, 1, 2),
	sC2("sC2", 2, 1, 2),
	sA3("sA3", 0, 2, 3),
	sB3("sB3", 1, 2, 3),
	sC3("sC3", 2, 2, 3),
	snk("snk", 3)
    {
	src.OutNumSentA.bind(NumSentA1);
	src.OutNumRcvdA.bind(NumRcvdA1);
	src.OutNumSentB.bind(NumSentB1);
	src.OutNumRcvdB.bind(NumRcvdB1);
	src.OutNumSentC.bind(NumSentC1);
	src.OutNumRcvdC.bind(NumRcvdC1);
	sA1.InNumSent.bind(NumSentA1);
	sA1.InNumRcvd.bind(NumRcvdA1);
	sA1.OutNumSent.bind(NumSentA2);
	sA1.OutNumRcvd.bind(NumRcvdA2);
	sA2.InNumSent.bind(NumSentA2);
	sA2.InNumRcvd.bind(NumRcvdA2);
	sA2.OutNumSent.bind(NumSentA3);
	sA2.OutNumRcvd.bind(NumRcvdA3);
	sA3.InNumSent.bind(NumSentA3);
	sA3.InNumRcvd.bind(NumRcvdA3);
	sA3.OutNumSent.bind(NumSentA4);
	sA3.OutNumRcvd.bind(NumRcvdA4);
	sB1.InNumSent.bind(NumSentB1);
	sB1.InNumRcvd.bind(NumRcvdB1);
	sB1.OutNumSent.bind(NumSentB2);
	sB1.OutNumRcvd.bind(NumRcvdB2);
	sB2.InNumSent.bind(NumSentB2);
	sB2.InNumRcvd.bind(NumRcvdB2);
	sB2.OutNumSent.bind(NumSentB3);
	sB2.OutNumRcvd.bind(NumRcvdB3);
	sB3.InNumSent.bind(NumSentB3);
	sB3.InNumRcvd.bind(NumRcvdB3);
	sB3.OutNumSent.bind(NumSentB4);
	sB3.OutNumRcvd.bind(NumRcvdB4);
	sC1.InNumSent.bind(NumSentC1);
	sC1.InNumRcvd.bind(NumRcvdC1);
	sC1.OutNumSent.bind(NumSentC2);
	sC1.OutNumRcvd.bind(NumRcvdC2);
	sC2.InNumSent.bind(NumSentC2);
	sC2.InNumRcvd.bind(NumRcvdC2);
	sC2.OutNumSent.bind(NumSentC3);
	sC2.OutNumRcvd.bind(NumRcvdC3);
	sC3.InNumSent.bind(NumSentC3);
	sC3.InNumRcvd.bind(NumRcvdC3);
	sC3.OutNumSent.bind(NumSentC4);
	sC3.OutNumRcvd.bind(NumRcvdC4);
	snk.InNumSentA.bind(NumSentA4);
	snk.InNumRcvdA.bind(NumRcvdA4);
	snk.InNumSentB.bind(NumSentB4);
	snk.InNumRcvdB.bind(NumRcvdB4);
	snk.InNumSentC.bind(NumSentC4);
	snk.InNumRcvdC.bind(NumRcvdC4);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("timed_pipe3x3_rnd2 starts\n");
/* double-check that unsigned counting works even if it wraps around **
	unsigned char a, b, c;
	a = 42; b = 42; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	a = 43; b = 42; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	a = 255; b = 255; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	a = 255; b = 254; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	a = 256; b = 255; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	a = 257; b = 256; c = a - b;
	printf("%u - %u = %u\n", a, b, c);
	exit(0);
** double-check that counting works even if it wraps around */
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    printf("timed_pipe3x3_rnd2 done.\n");
    printf("[potential speedup over SEQ: OOO 9x, NPD 9x, SYN 9x]\n");
    return 0;
}

// EOF
