// timed_pipe9_rnd.cc: timed pipeline with 9 inner stages and random workload
// (inspired from delta_pipe9_rnd.cc)
// 05/02/20, RD

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define BUFSIZE    2
#define STAGES     9
#define ITEMS      10
#define WORKLOAD   10000000*STAGES
#define WORKTIME   5
//#define VERBOSE

#pragma RISC M conflict-free
int M[STAGES+1][BUFSIZE] = { }; // memory for double-buffering

int Rnd(unsigned long &Next, int Max)
{
    Next = Next * 1103515245 + 12345;
    return((unsigned)(Next/65536) % (Max+1));
}

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

void WaitAnyTime(unsigned long &RndNext, int MaxPicos)
{
    int picos = Rnd(RndNext, MaxPicos+1)-1; // range [-1...MaxPicos]

    if(picos >= 0)	// don't wait if picos is negative;
    {	// wait for a delta-cycle or some time
//	printf("DEBUG WaitAnyTime waiting for %d pico seconds.\n", picos);
	wait((double)picos, SC_PS);
    }
    else
    {	// skip (no wait)
//	printf("DEBUG WaitAnyTime NOT waiting.\n");
    }
}

SC_MODULE(Source)
{
    int Buffer, Index, Count;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> OutNumSent;

    void main(void)
    {
	unsigned long RndNext = 0;

	for(int i=0; i<ITEMS; i++)
	{
	    // make an item
	    WaitAnyTime(RndNext, WORKTIME);
	    Count++;
	    WaitAnyTime(RndNext, WORKTIME);

	    // if buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Source waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in buffer and notify receiver
	    M[Buffer][Index] = Count;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), Count);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next send
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*ITEMS));
    }

    Source(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Source);
	SC_THREAD(main);
    }
};

SC_MODULE(Stage)
{
    int InBuffer, OutBuffer, Index, d;
    sc_in<unsigned>  InNumSent;
    sc_in<unsigned>  OutNumRcvd;
    sc_out<unsigned> InNumRcvd;
    sc_out<unsigned> OutNumSent;
    unsigned long w;

    void main(void)
    {
	unsigned long RndNext = InBuffer;

	for(;;)
	{
	    // if input buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for input buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load item from buffer and notify sender
	    d = M[InBuffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction

	    // process the item
#ifdef VERBOSE
	    printf("%2u:%02u: %s processes data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    WaitAnyTime(RndNext, WORKTIME);
	    w = Work(WORKLOAD);
	    WaitAnyTime(RndNext, WORKTIME);

	    // if output buffer is full, wait for free space
	    while (OutNumSent - OutNumRcvd >= BUFSIZE)
	    {
//		printf("%2u:%02u: DEBUG Stage waiting for output buffer space S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			OutNumSent - OutNumRcvd);
		wait(OutNumRcvd.value_changed_event());
	    }

	    // store item in output buffer and notify receiver
	    M[OutBuffer][Index] = d;
	    OutNumSent = OutNumSent + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s sent data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next transaction
	}
    }

    Stage(sc_module_name _nm, int _ib, int _ob)
    :	sc_module(_nm), InBuffer(_ib), OutBuffer(_ob)
    {
	Index = 0;
	SC_HAS_PROCESS(Stage);
	SC_THREAD(main);
    }
};

SC_MODULE(Sink)
{
    int Buffer, Index, d, Count;
    sc_in<unsigned>  InNumSent;
    sc_out<unsigned> InNumRcvd;

    void main(void)
    {
	unsigned long RndNext = 1;

	for(int i=0; i<ITEMS; i++)
	{
	    // if buffer is empty, wait for items
	    while (InNumSent - InNumRcvd == 0)
	    {
//		printf("%2u:%02u: DEBUG Sink waiting for item S=%u\n",
//			(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
//			InNumSent - InNumRcvd);
		wait(InNumSent.value_changed_event());
	    }

	    // load items from buffer and notify sender
	    d = M[Buffer][Index];
	    InNumRcvd = InNumRcvd + 1;
#ifdef VERBOSE
	    printf("%2u:%02u: %s received data %d\n",
		(unsigned)sc_time_stamp().value(), (unsigned)sc_delta_count(),
		name(), d);
#endif
	    Index = (Index + 1) % BUFSIZE;
	    wait(SC_ZERO_TIME); // must wait (for _some_ time) before next receive

	    // check the item
	    WaitAnyTime(RndNext, WORKTIME);
	    Count++;
	    assert(d == Count);
	    WaitAnyTime(RndNext, WORKTIME);
	}
	assert(Count == ITEMS);
	assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    }

    Sink(sc_module_name _nm, int _b)
    :	sc_module(_nm), Buffer(_b)
    {
	Index = 0;
	Count = 0;
	SC_HAS_PROCESS(Sink);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    sc_signal<unsigned> NumSent1, NumRcvd1;
    sc_signal<unsigned> NumSent2, NumRcvd2;
    sc_signal<unsigned> NumSent3, NumRcvd3;
    sc_signal<unsigned> NumSent4, NumRcvd4;
    sc_signal<unsigned> NumSent5, NumRcvd5;
    sc_signal<unsigned> NumSent6, NumRcvd6;
    sc_signal<unsigned> NumSent7, NumRcvd7;
    sc_signal<unsigned> NumSent8, NumRcvd8;
    sc_signal<unsigned> NumSent9, NumRcvd9;
    sc_signal<unsigned> NumSent10, NumRcvd10;

    Source src;
    Stage  s1, s2, s3, s4, s5, s6, s7, s8, s9;
    Sink   snk;

    SC_CTOR(TOP)
    : NumSent1("NumSent1", 0),
	NumRcvd1("NumRcvd1", 0),
	NumSent2("NumSent2", 0),
	NumRcvd2("NumRcvd2", 0),
	NumSent3("NumSent3", 0),
	NumRcvd3("NumRcvd3", 0),
	NumSent4("NumSent4", 0),
	NumRcvd4("NumRcvd4", 0),
	NumSent5("NumSent5", 0),
	NumRcvd5("NumRcvd5", 0),
	NumSent6("NumSent6", 0),
	NumRcvd6("NumRcvd6", 0),
	NumSent7("NumSent7", 0),
	NumRcvd7("NumRcvd7", 0),
	NumSent8("NumSent8", 0),
	NumRcvd8("NumRcvd8", 0),
	NumSent9("NumSent9", 0),
	NumRcvd9("NumRcvd9", 0),
	NumSent10("NumSent10", 0),
	NumRcvd10("NumRcvd10", 0),
	src("src", 0),
	s1("s1", 0, 1),
	s2("s2", 1, 2),
	s3("s3", 2, 3),
	s4("s4", 3, 4),
	s5("s5", 4, 5),
	s6("s6", 5, 6),
	s7("s7", 6, 7),
	s8("s8", 7, 8),
	s9("s9", 8, 9),
	snk("snk", 9)
    {
	src.OutNumSent.bind(NumSent1);
	src.OutNumRcvd.bind(NumRcvd1);
	s1.InNumSent.bind(NumSent1);
	s1.InNumRcvd.bind(NumRcvd1);
	s1.OutNumSent.bind(NumSent2);
	s1.OutNumRcvd.bind(NumRcvd2);
	s2.InNumSent.bind(NumSent2);
	s2.InNumRcvd.bind(NumRcvd2);
	s2.OutNumSent.bind(NumSent3);
	s2.OutNumRcvd.bind(NumRcvd3);
	s3.InNumSent.bind(NumSent3);
	s3.InNumRcvd.bind(NumRcvd3);
	s3.OutNumSent.bind(NumSent4);
	s3.OutNumRcvd.bind(NumRcvd4);
	s4.InNumSent.bind(NumSent4);
	s4.InNumRcvd.bind(NumRcvd4);
	s4.OutNumSent.bind(NumSent5);
	s4.OutNumRcvd.bind(NumRcvd5);
	s5.InNumSent.bind(NumSent5);
	s5.InNumRcvd.bind(NumRcvd5);
	s5.OutNumSent.bind(NumSent6);
	s5.OutNumRcvd.bind(NumRcvd6);
	s6.InNumSent.bind(NumSent6);
	s6.InNumRcvd.bind(NumRcvd6);
	s6.OutNumSent.bind(NumSent7);
	s6.OutNumRcvd.bind(NumRcvd7);
	s7.InNumSent.bind(NumSent7);
	s7.InNumRcvd.bind(NumRcvd7);
	s7.OutNumSent.bind(NumSent8);
	s7.OutNumRcvd.bind(NumRcvd8);
	s8.InNumSent.bind(NumSent8);
	s8.InNumRcvd.bind(NumRcvd8);
	s8.OutNumSent.bind(NumSent9);
	s8.OutNumRcvd.bind(NumRcvd9);
	s9.InNumSent.bind(NumSent9);
	s9.InNumRcvd.bind(NumRcvd9);
	s9.OutNumSent.bind(NumSent10);
	s9.OutNumRcvd.bind(NumRcvd10);
	snk.InNumSent.bind(NumSent10);
	snk.InNumRcvd.bind(NumRcvd10);
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("timed_pipe9_rnd starts\n");
    sc_start();
    assert(top.src.Count == ITEMS);
    assert(top.snk.Count == ITEMS);
    assert(sc_time_stamp().value() <= (unsigned)(2*WORKTIME*(ITEMS+STAGES+1)));
    printf("timed_pipe9_rnd done.\n");
    printf("[potential speedup over SEQ: OOO 9x, NPD 9x, SYN 9x]\n");
    return 0;
}

// EOF
